import ee
from ee_extra.JavaScript.install import install as ee_install
from ee_extra.JavaScript.install import uninstall as ee_uninstall
from ee_extra.JavaScript.main import ee_require

from .extending import extend


@extend(ee)
def require(module):
    """Loads and executes a JavaScript GEE module.

    All modules must be first installed before requiring them. After requiring the module,
    it can be used in the same way as it is used in the Code Editor.

    Warning
    -------
    This method is highly :code:`experimental`. Please report any irregularities in the
    Issues Page of `eeExtra <https://github.com/r-earthengine/ee_extra>`_.

    Parameters
    ----------
    module : str
        Path to the module in the Code Editor (e.g. "users/dmlmont/spectral:spectral").

    Returns
    -------
    BoxDict
        Loaded module. Methods and attributes can be accessed using dot notation.

    See Also
    --------
    install : Installs a JavaScript GEE module.
    uninstall : Uninstalls a JavaScript GEE module.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> LandsatLST = ee.require("users/sofiaermida/landsat_smw_lst:modules/Landsat_LST.js")
    """
    return ee_require(module)


@extend(ee)
def install(module, update=False, quiet=False):
    """Installs a JavaScript GEE module.

    Warning
    -------
    This method is highly :code:`experimental`. Please report any irregularities in the
    Issues Page of `eeExtra <https://github.com/r-earthengine/ee_extra>`_.

    Parameters
    ----------
    module : str
        Path to the module in the Code Editor (e.g. "users/dmlmont/spectral:spectral").
    update : bool, default = False
        Whether to update the module if it is already installed.
    quiet : bool, default = False
        Whether to show in console the process.

    Returns
    -------
    None

    See Also
    --------
    uninstall : Uninstalls a JavaScript GEE module.
    require : Loads and executes a JavaScript GEE module.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.install("users/sofiaermida/landsat_smw_lst:modules/Landsat_LST.js")
    """
    return ee_install(module, update, quiet)


@extend(ee)
def uninstall(module, quiet=False):
    """Uninstalls a JavaScript GEE module.

    Warning
    -------
    This method is highly :code:`experimental`. Please report any irregularities in the
    Issues Page of `eeExtra <https://github.com/r-earthengine/ee_extra>`_.

    Parameters
    ----------
    module : str
        Path to the module in the Code Editor (e.g. "users/dmlmont/spectral:spectral").
    quiet : bool, default = False
        Whether to show in console the process.

    Returns
    -------
    None

    See Also
    --------
    install : Installs a JavaScript GEE module.
    require : Loads and executes a JavaScript GEE module.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.uninstall("users/sofiaermida/landsat_smw_lst:modules/Landsat_LST.js")
    """
    return ee_uninstall(module, quiet)



import warnings

import ee
import ee_extra
import ee_extra.Spectral.core
import ee_extra.Algorithms.core
import requests

from .extending import extend


@extend(ee.image.Image)
def __getitem__(self, key):
    """Gets the band of the image according to the specified key.

    Parameters
    ----------
    self : ee.Image
        Image to get the bands from.
    key : numeric | string | list[numeric] | list[string] | slice
        Key used to get the specified band. If numeric, it gets the band at that index.
        If string, it gets the band with that name or that matches with regex. If list, it
        gets multiple bands. If slice, it calls the slice() method (the step parameter is
        ignored).

    Returns
    -------
    ee.Image
        Image with the selected bands.
    """
    if isinstance(key, slice):

        if key.start == None:
            start = 0
        else:
            start = key.start

        if key.stop == None:
            stop = self.bandNames().size()
        else:
            stop = key.stop

        selected = self.slice(start, stop)

    else:
        selected = self.select(key)

    return selected


@extend(ee.image.Image)
def __add__(self, other):
    """Computes the addition between two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Addition of two images.
    """
    return self.add(other)


@extend(ee.image.Image)
def __radd__(self, other):
    """Computes the addition between two images.

    Parameters
    ----------
    self : ee.Image
        Right operand.
    other : ee.Image | numeric | list[numeric]
        Left operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Addition of two images.
    """
    return self.add(other)


@extend(ee.image.Image)
def __sub__(self, other):
    """Computes the subtraction between two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Subtraction of two images.
    """
    return self.subtract(other)


@extend(ee.image.Image)
def __rsub__(self, other):
    """Computes the subtraction between two images.

    Parameters
    ----------
    self : ee.Image
        Right operand.
    other : ee.Image | numeric | list[numeric]
        Left operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Subtraction of two images.
    """
    return ee.Image(other).subtract(self)


@extend(ee.image.Image)
def __mul__(self, other):
    """Computes the multiplication between two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Multiplication of two images.
    """
    return self.multiply(other)


@extend(ee.image.Image)
def __rmul__(self, other):
    """Computes the multiplication between two images.

    Parameters
    ----------
    self : ee.Image
        Right operand.
    other : ee.Image | numeric | list[numeric]
        Left operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Multiplication of two images.
    """
    return self.multiply(other)


@extend(ee.image.Image)
def __truediv__(self, other):
    """Computes the division between two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Division of two images.
    """
    return self.divide(other)


@extend(ee.image.Image)
def __rtruediv__(self, other):
    """Computes the division between two images.

    Parameters
    ----------
    self : ee.Image
        Right operand.
    other : ee.Image | numeric | list[numeric]
        Left operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Division of two images.
    """
    return ee.Image(other).divide(self)


@extend(ee.image.Image)
def __floordiv__(self, other):
    """Computes the floor division of two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Floor division of two images.
    """
    return self.divide(other).floor()


@extend(ee.image.Image)
def __rfloordiv__(self, other):
    """Computes the floor division of two images.

    Parameters
    ----------
    self : ee.Image
        Right operand.
    other : ee.Image | numeric | list[numeric]
        Left operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Floor division of two images.
    """
    return ee.Image(other).divide(self).floor()


@extend(ee.image.Image)
def __mod__(self, other):
    """Computes the modulo of two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Modulo of two images.
    """
    return self.mod(other)


@extend(ee.image.Image)
def __rmod__(self, other):
    """Computes the modulo of two images.

    Parameters
    ----------
    self : ee.Image
        Right operand.
    other : ee.Image | numeric | list[numeric]
        Left operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Modulo of two images.
    """
    return ee.Image(other).mod(self)


@extend(ee.image.Image)
def __pow__(self, other):
    """Computes the base (left operand) to the power (right operand).

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Bsae to the power of two images.
    """
    return self.pow(other)


@extend(ee.image.Image)
def __rpow__(self, other):
    """Computes the base (left operand) to the power (right operand).

    Parameters
    ----------
    self : ee.Image
        Right operand.
    other : ee.Image | numeric | list[numeric]
        Left operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Base to the power of two images.
    """
    return ee.Image(other).pow(self)


@extend(ee.image.Image)
def __lshift__(self, other):
    """Computes the left shift operation between two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Left shift operation.
    """
    return self.leftShift(other)


@extend(ee.image.Image)
def __rlshift__(self, other):
    """Computes the left shift operation between two images.

    Parameters
    ----------
    self : ee.Image
        Right operand.
    other : ee.Image | numeric | list[numeric]
        Left operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Left shift operation.
    """
    return ee.Image(other).leftShift(self)


@extend(ee.image.Image)
def __rshift__(self, other):
    """Computes the right shift operation between two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Right shift operation.
    """
    return self.rightShift(other)


@extend(ee.image.Image)
def __rrshift__(self, other):
    """Computes the right shift operation between two images.

    Parameters
    ----------
    self : ee.Image
        Right operand.
    other : ee.Image | numeric | list[numeric]
        Left operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Right shift operation.
    """
    return ee.Image(other).rightShift(self)


@extend(ee.image.Image)
def __and__(self, other):
    """Computes the binary operator AND between two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Binary operator AND.
    """
    return self.And(other)


@extend(ee.image.Image)
def __rand__(self, other):
    """Computes the binary operator AND between two images.

    Parameters
    ----------
    self : ee.Image
        Right operand.
    other : ee.Image | numeric | list[numeric]
        Left operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Binary operator AND.
    """
    return ee.Image(other).And(self)


@extend(ee.image.Image)
def __or__(self, other):
    """Computes the binary operator OR between two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Binary operator OR.
    """
    return self.Or(other)


@extend(ee.image.Image)
def __ror__(self, other):
    """Computes the binary operator OR between two images.

    Parameters
    ----------
    self : ee.Image
        Right operand.
    other : ee.Image | numeric | list[numeric]
        Left operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Binary operator OR.
    """
    return ee.Image(other).Or(self)


@extend(ee.image.Image)
def __lt__(self, other):
    """Computes the rich comparison LOWER THAN between two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Rich comparison LOWER THAN.
    """
    return self.lt(other)


@extend(ee.image.Image)
def __le__(self, other):
    """Computes the rich comparison LOWER THAN OR EQUAL between two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Rich comparison LOWER THAN OR EQUAL.
    """
    return self.lte(other)


@extend(ee.image.Image)
def __eq__(self, other):
    """Computes the rich comparison EQUAL between two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Rich comparison EQUAL.
    """
    return self.eq(other)


@extend(ee.image.Image)
def __ne__(self, other):
    """Computes the rich comparison NOT EQUAL THAN between two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Rich comparison NOT EQUAL.
    """
    return self.neq(other)


@extend(ee.image.Image)
def __gt__(self, other):
    """Computes the rich comparison GREATER THAN between two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Rich comparison GREATER THAN.
    """
    return self.gt(other)


@extend(ee.image.Image)
def __ge__(self, other):
    """Computes the rich comparison GREATER THAN OR EQUAL between two images.

    Parameters
    ----------
    self : ee.Image
        Left operand.
    other : ee.Image | numeric | list[numeric]
        Right operand. If numeric, an ee.Image is created from its value. If list, an
        ee.Image with n bands (n = len(list)) is created from its values.

    Returns
    -------
    ee.Image
        Rich comparison GREATER THAN OR EQUAL.
    """
    return self.gte(other)


@extend(ee.image.Image)
def __neg__(self):
    """Computes the unary operator NEGATIVE on an image.

    Parameters
    ----------
    self : ee.Image
        Operand.

    Returns
    -------
    ee.Image
        Unary operator NEGATIVE.
    """
    return self.multiply(-1)


@extend(ee.image.Image)
def __invert__(self):
    """Computes the unary operator NOT on an image.

    Parameters
    ----------
    self : ee.Image
        Operand.

    Returns
    -------
    ee.Image
        Unary operator NOT.
    """
    return self.Not()


@extend(ee.image.Image)
def index(
    self,
    index="NDVI",
    G=2.5,
    C1=6.0,
    C2=7.5,
    L=1.0,
    cexp=1.16,
    nexp=2.0,
    alpha=0.1,
    slope=1.0,
    intercept=0.0,
    gamma=1.0,
    kernel="RBF",
    sigma="0.5 * (a + b)",
    p=2.0,
    c=1.0,
    online=False,
    drop=False,
):
    """Computes one or more spectral indices (indices are added as bands) for an image.

    .. deprecated:: 0.3.0
       Use :func:`spectralIndices()` instead.

    Tip
    ----------
    Check more info about the supported platforms and spectral indices in the
    :ref:`User Guide<Spectral Indices Computation>`.

    Parameters
    ----------
    self : ee.Image [this]
        Image to compute indices on. Must be scaled to [0,1].
    index : string | list[string], default = 'NDVI'
        Index or list of indices to compute.\n
        Available options:
            - 'vegetation' : Compute all vegetation indices.
            - 'burn' : Compute all burn indices.
            - 'water' : Compute all water indices.
            - 'snow' : Compute all snow indices.
            - 'drought' : Compute all drought indices.
            - 'urban' : Compute all urban (built-up) indices.
            - 'kernel' : Compute all kernel indices.
            - 'all' : Compute all indices listed below.
        Awesome Spectral Indices for GEE:
            Check the complete list of indices
            `here <https://awesome-ee-spectral-indices.readthedocs.io/en/latest/list.html>`_.
    G : float, default = 2.5
        Gain factor. Used just for index = 'EVI'.
    C1 : float, default = 6.0
        Coefficient 1 for the aerosol resistance term. Used just for index = 'EVI'.
    C2 : float, default = 7.5
        Coefficient 2 for the aerosol resistance term. Used just for index = 'EVI'.
    L : float, default = 1.0
        Canopy background adjustment. Used just for index = ['EVI','SAVI'].
    cexp : float, default = 1.16
        Exponent used for OCVI.
    nexp : float, default = 2.0
        Exponent used for GDVI.
    alpha : float, default = 0.1
        Weighting coefficient used for WDRVI.
    slope : float, default = 1.0
        Soil line slope.
    intercept : float, default = 0.0
        Soil line intercept.
    gamma : float, default = 1.0
        Weighting coefficient used for ARVI.
    kernel : str, default = 'RBF'
        Kernel used for kernel indices.\n
        Available options:
            - 'linear' : Linear Kernel.
            - 'RBF' : Radial Basis Function (RBF) Kernel.
            - 'poly' : Polynomial Kernel.
    sigma : str | float, default = '0.5 * (a + b)'
        Length-scale parameter. Used for kernel = 'RBF'. If str, this must be an
        expression including 'a' and 'b'. If numeric, this must be positive.
    p : float, default = 2.0
        Kernel degree. Used for kernel = 'poly'.
    c : float, default = 1.0
        Free parameter that trades off the influence of higher-order versus lower-order
        terms in the polynomial kernel. Used for kernel = 'poly'. This must be greater
        than or equal to 0.
    online : boolean, default = False
        Whether to retrieve the most recent list of indices directly from the GitHub
        repository and not from the local copy.
    drop : boolean, default = True
        Whether to drop all bands except the new spectral indices.

    Returns
    -------
    ee.Image
        Image with the computed spectral index, or indices, as new bands.

    See Also
    --------
    scale : Scales bands on an image collection.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> S2 = ee.ImageCollection('COPERNICUS/S2_SR').scale().first()

    - Computing one spectral index:

    >>> S2.index('NDVI')

    - Computing indices with different parameters:

    >>> S2.index('SAVI',L = 0.5)

    - Computing multiple indices:

    >>> S2.index(['NDVI','EVI','GNDVI'])

    - Computing a specific group of indices:

    >>> S2.index('vegetation')

    - Computing kernel indices:

    >>> S2.index(['kNDVI'],kernel = 'poly',p = 5)

    - Computing all indices:

    >>> S2.index('all')
    """
    warnings.warn(
        "index() is deprecated, please use spectralIndices() instead",
        DeprecationWarning,
    )

    return self


@extend(ee.image.Image)
def spectralIndices(
    self,
    index="NDVI",
    G=2.5,
    C1=6.0,
    C2=7.5,
    L=1.0,
    cexp=1.16,
    nexp=2.0,
    alpha=0.1,
    slope=1.0,
    intercept=0.0,
    gamma=1.0,
    omega=2.0,
    beta=0.05,
    k=0.0,
    kernel="RBF",
    sigma="0.5 * (a + b)",
    p=2.0,
    c=1.0,
    lambdaN=858.5,
    lambdaR=645.0,
    lambdaG=555.0,
    online=False,
    drop=False,
):
    """Computes one or more spectral indices (indices are added as bands) for an image
    from the Awesome List of Spectral Indices.

    Tip
    ----------
    Check more info about the supported platforms and spectral indices in the
    :ref:`User Guide<Spectral Indices Computation>`.

    Parameters
    ----------
    self : ee.Image [this]
        Image to compute indices on. Must be scaled to [0,1].
    index : string | list[string], default = 'NDVI'
        Index or list of indices to compute.\n
        Available options:
            - 'vegetation' : Compute all vegetation indices.
            - 'burn' : Compute all burn indices.
            - 'water' : Compute all water indices.
            - 'snow' : Compute all snow indices.
            - 'urban' : Compute all urban (built-up) indices.
            - 'kernel' : Compute all kernel indices.
            - 'all' : Compute all indices listed below.
        Awesome Spectral Indices for GEE:
            Check the complete list of indices
            `here <https://awesome-ee-spectral-indices.readthedocs.io/en/latest/list.html>`_.
    G : float, default = 2.5
        Gain factor. Used just for index = 'EVI'.
    C1 : float, default = 6.0
        Coefficient 1 for the aerosol resistance term. Used just for index = 'EVI'.
    C2 : float, default = 7.5
        Coefficient 2 for the aerosol resistance term. Used just for index = 'EVI'.
    L : float, default = 1.0
        Canopy background adjustment. Used just for index = ['EVI','SAVI'].
    cexp : float, default = 1.16
        Exponent used for OCVI.
    nexp : float, default = 2.0
        Exponent used for GDVI.
    alpha : float, default = 0.1
        Weighting coefficient used for WDRVI.
    slope : float, default = 1.0
        Soil line slope.
    intercept : float, default = 0.0
        Soil line intercept.
    gamma : float, default = 1.0
        Weighting coefficient used for ARVI.
    omega : float, default = 2.0
        Weighting coefficient  used for MBWI.
    beta : float, default = 0.05
        Calibration parameter used for NDSIns.
    k : float, default = 0.0
        Slope parameter by soil used for NIRvH2.
    kernel : str, default = 'RBF'
        Kernel used for kernel indices.\n
        Available options:
            - 'linear' : Linear Kernel.
            - 'RBF' : Radial Basis Function (RBF) Kernel.
            - 'poly' : Polynomial Kernel.
    sigma : str | float, default = '0.5 * (a + b)'
        Length-scale parameter. Used for kernel = 'RBF'. If str, this must be an
        expression including 'a' and 'b'. If numeric, this must be positive.
    p : float, default = 2.0
        Kernel degree. Used for kernel = 'poly'.
    c : float, default = 1.0
        Free parameter that trades off the influence of higher-order versus lower-order
        terms in the polynomial kernel. Used for kernel = 'poly'. This must be greater
        than or equal to 0.
    lambdaN : float, default = 858.5
        NIR wavelength used for NIRvH2 and NDGI.
    lambdaR : float, default = 645.0
        Red wavelength used for NIRvH2 and NDGI.
    lambdaG : float, default = 555.0
        Green wavelength used for NDGI.
    online : boolean, default = False
        Whether to retrieve the most recent list of indices directly from the GitHub
        repository and not from the local copy.
    drop : boolean, default = True
        Whether to drop all bands except the new spectral indices.

    Returns
    -------
    ee.Image
        Image with the computed spectral index, or indices, as new bands.

    See Also
    --------
    scaleAndOffset : Scales bands on an image collection.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> S2 = ee.ImageCollection('COPERNICUS/S2_SR').scaleAndOffset().first()

    - Computing one spectral index:

    >>> S2.spectralIndices('NDVI')

    - Computing indices with different parameters:

    >>> S2.spectralIndices('SAVI',L = 0.5)

    - Computing multiple indices:

    >>> S2.spectralIndices(['NDVI','EVI','GNDVI'])

    - Computing a specific group of indices:

    >>> S2.spectralIndices('vegetation')

    - Computing kernel indices:

    >>> S2.spectralIndices(['kNDVI'],kernel = 'poly',p = 5)

    - Computing all indices:

    >>> S2.spectralIndices('all')
    """
    return ee_extra.Spectral.core.spectralIndices(
        self,
        index,
        G,
        C1,
        C2,
        L,
        cexp,
        nexp,
        alpha,
        slope,
        intercept,
        gamma,
        omega,
        beta,
        k,
        kernel,
        sigma,
        p,
        c,
        lambdaN,
        lambdaR,
        lambdaG,
        online,
        drop,
    )


@extend(ee.image.Image)
def maskClouds(
    self,
    method="cloud_prob",
    prob=60,
    maskCirrus=True,
    maskShadows=True,
    scaledImage=False,
    dark=0.15,
    cloudDist=1000,
    buffer=250,
    cdi=None,
):
    """Masks clouds and shadows in an image (valid just for Surface Reflectance products).

    Tip
    ----------
    Check more info about the supported platforms and clouds masking in the
    :ref:`User Guide<Masking Clouds and Shadows>`.

    Parameters
    ----------
    self : ee.Image [this]
        Image to mask.
    method : string, default = 'cloud_prob'
        Method used to mask clouds.\n
        Available options:
            - 'cloud_prob' : Use cloud probability.
            - 'qa' : Use Quality Assessment band.
        This parameter is ignored for Landsat products.
    prob : numeric [0, 100], default = 60
        Cloud probability threshold. Valid just for method = 'cloud_prob'. This parameter
        is ignored for Landsat products.
    maskCirrus : boolean, default = True
        Whether to mask cirrus clouds. Valid just for method = 'qa'. This parameter is
        ignored for Landsat products.
    maskShadows : boolean, default = True
        Whether to mask cloud shadows. For more info see 'Braaten, J. 2020. Sentinel-2
        Cloud Masking with s2cloudless. Google Earth Engine, Community Tutorials'.
    scaledImage : boolean, default = False
        Whether the pixel values are scaled to the range [0,1] (reflectance values).
        This parameter is ignored for Landsat products.
    dark : float [0,1], default = 0.15
        NIR threshold. NIR values below this threshold are potential cloud shadows.
        This parameter is ignored for Landsat products.
    cloudDist : int, default = 1000
        Maximum distance in meters (m) to look for cloud shadows from cloud edges.
        This parameter is ignored for Landsat products.
    buffer : int, default = 250
        Distance in meters (m) to dilate cloud and cloud shadows objects. This parameter
        is ignored for Landsat products.
    cdi : float [-1,1], default = None
        Cloud Displacement Index threshold. Values below this threshold are considered
        potential clouds. A cdi = None means that the index is not used. For more info see
        'Frantz, D., HaS, E., Uhl, A., Stoffels, J., Hill, J. 2018. Improvement of the
        Fmask algorithm for Sentinel-2 images: Separating clouds from bright surfaces
        based on parallax effects. Remote Sensing of Environment 2015: 471-481'.
        This parameter is ignored for Landsat products.

    Returns
    -------
    ee.Image
        Cloud-shadow masked image.

    Notes
    -----
    This method may mask water as well as clouds for the Sentinel-3 Radiance product.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> S2 = (ee.ImageCollection('COPERNICUS/S2_SR')
    ...     .first()
    ...     .maskClouds(prob = 75,buffer = 300,cdi = -0.5))
    """
    return ee_extra.QA.clouds.maskClouds(
        self,
        method,
        prob,
        maskCirrus,
        maskShadows,
        scaledImage,
        dark,
        cloudDist,
        buffer,
        cdi,
    )


@extend(ee.image.Image)
def scale(self):
    """Scales bands on an image.

    .. deprecated:: 0.3.0
       Use :func:`scaleAndOffset()` instead.

    Tip
    ----------
    Check more info about the supported platforms and image scaling the
    :ref:`User Guide<Image Scaling>`.

    Parameters
    ----------
    self : ee.Image [this]
        Image to scale.

    Returns
    -------
    ee.Image
        Scaled image.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> S2 = ee.ImageCollection('COPERNICUS/S2_SR').first().scale()
    """
    warnings.warn(
        "scale() is deprecated, please use scaleAndOffset() instead",
        DeprecationWarning,
    )

    return ee_extra.STAC.core.scaleAndOffset(self)


@extend(ee.image.Image)
def getScaleParams(self):
    """Gets the scale parameters for each band of the image.

    Parameters
    ----------
    self : ee.Image (this)
        Image to get the scale parameters from.

    Returns
    -------
    dict
        Dictionary with the scale parameters for each band.

    See Also
    --------
    getOffsetParams : Gets the offset parameters for each band of the image.
    scaleAndOffset : Scales bands on an image according to their scale and offset parameters.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.ImageCollection('MODIS/006/MOD11A2').first().getScaleParams()
    {'Clear_sky_days': 1.0,
     'Clear_sky_nights': 1.0,
     'Day_view_angl': 1.0,
     'Day_view_time': 0.1,
     'Emis_31': 0.002,
     'Emis_32': 0.002,
     'LST_Day_1km': 0.02,
     'LST_Night_1km': 0.02,
     'Night_view_angl': 1.0,
     'Night_view_time': 0.1,
     'QC_Day': 1.0,
     'QC_Night': 1.0}
    """
    return ee_extra.STAC.core.getScaleParams(self)


@extend(ee.image.Image)
def getOffsetParams(self):
    """Gets the offset parameters for each band of the image.

    Parameters
    ----------
    self : ee.Image (this)
        Image to get the offset parameters from.

    Returns
    -------
    dict
        Dictionary with the offset parameters for each band.

    See Also
    --------
    getScaleParams : Gets the scale parameters for each band of the image.
    scaleAndOffset : Scales bands on an image according to their scale and offset parameters.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.ImageCollection('MODIS/006/MOD11A2').first().getOffsetParams()
    {'Clear_sky_days': 0.0,
     'Clear_sky_nights': 0.0,
     'Day_view_angl': -65.0,
     'Day_view_time': 0.0,
     'Emis_31': 0.49,
     'Emis_32': 0.49,
     'LST_Day_1km': 0.0,
     'LST_Night_1km': 0.0,
     'Night_view_angl': -65.0,
     'Night_view_time': 0.0,
     'QC_Day': 0.0,
     'QC_Night': 0.0}
    """
    return ee_extra.STAC.core.getOffsetParams(self)


@extend(ee.image.Image)
def scaleAndOffset(self):
    """Scales bands on an image according to their scale and offset parameters.

    Tip
    ----------
    Check more info about the supported platforms and image scaling the
    :ref:`User Guide<Image Scaling>`.

    Parameters
    ----------
    self : ee.Image [this]
        Image to scale.

    Returns
    -------
    ee.Image
        Scaled image.

    See Also
    --------
    getScaleParams : Gets the scale parameters for each band of the image.
    getOffsetParams : Gets the offset parameters for each band of the image.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> S2 = ee.ImageCollection('COPERNICUS/S2_SR').first().scaleAndOffset()
    """
    return ee_extra.STAC.core.scaleAndOffset(self)


@extend(ee.image.Image)
def preprocess(self, **kwargs):
    """Pre-processes the image: masks clouds and shadows, and scales and offsets the image.

    Tip
    ----------
    Check more info here about the supported platforms, :ref:`Image Scaling<Image Scaling>`
    and :ref:`Masking Clouds and Shadows<Masking Clouds and Shadows>`.

    Parameters
    ----------
    self : ee.Image [this]
        Image to pre-process.
    **kwargs :
        Keywords arguments for maskClouds().

    Returns
    -------
    ee.Image
        Pre-processed image.

    See Also
    --------
    getScaleParams : Gets the scale parameters for each band of the image.
    getOffsetParams : Gets the offset parameters for each band of the image.
    scaleAndOffset : Scales bands on an image according to their scale and offset parameters.
    maskClouds : Masks clouds and shadows in an image.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> S2 = ee.ImageCollection('COPERNICUS/S2_SR').first().preprocess()
    """
    return ee_extra.QA.pipelines.preprocess(self, **kwargs)


@extend(ee.image.Image)
def getSTAC(self):
    """Gets the STAC of the image.

    Parameters
    ----------
    self : ee.Image [this]
        Image to get the STAC from.

    Returns
    -------
    dict
        STAC of the image.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.ImageCollection('COPERNICUS/S2_SR').first().getSTAC()
    {'stac_version': '1.0.0-rc.2',
     'type': 'Collection',
     'stac_extensions': ['https://stac-extensions.github.io/eo/v1.0.0/schema.json'],
     'id': 'COPERNICUS/S2_SR',
     'title': 'Sentinel-2 MSI: MultiSpectral Instrument, Level-2A',
     'gee:type': 'image_collection',
     ...}
    """
    return ee_extra.STAC.core.getSTAC(self)


@extend(ee.image.Image)
def getDOI(self):
    """Gets the DOI of the image, if available.

    Parameters
    ----------
    self : ee.Image [this]
        Image to get the DOI from.

    Returns
    -------
    str
        DOI of the ee.Image dataset.

    See Also
    --------
    getCitation : Gets the citation of the image, if available.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.ImageCollection('NASA/GPM_L3/IMERG_V06').first().getDOI()
    '10.5067/GPM/IMERG/3B-HH/06'
    """
    return ee_extra.STAC.core.getDOI(self)


@extend(ee.image.Image)
def getCitation(self):
    """Gets the citation of the image, if available.

    Parameters
    ----------
    self : ee.Image [this]
        Image to get the citation from.

    Returns
    -------
    str
        Citation of the ee.Image dataset.

    See Also
    --------
    getDOI : Gets the DOI of the image, if available.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.ImageCollection('NASA/GPM_L3/IMERG_V06').first().getCitation()
    'Huffman, G.J., E.F. Stocker, D.T. Bolvin, E.J. Nelkin, Jackson Tan (2019),
    GPM IMERG Final Precipitation L3 Half Hourly 0.1 degree x 0.1 degree V06, Greenbelt,
    MD, Goddard Earth Sciences Data and Information Services Center (GES DISC),
    Accessed: [Data Access Date],
    [doi:10.5067/GPM/IMERG/3B-HH/06](https://doi.org/10.5067/GPM/IMERG/3B-HH/06)'
    """
    return ee_extra.STAC.core.getCitation(self)


@extend(ee.image.Image)
def panSharpen(self, method="SFIM", qa=None, **kwargs):
    """Apply panchromatic sharpening to the Image.

    Optionally, run quality assessments between the original and sharpened Image to
    measure spectral distortion and set results as properties of the sharpened Image.

    Parameters
    ----------
    self : ee.Image [this]
        Image to sharpen.
    method : str, default="SFIM"
        The sharpening algorithm to apply. Current options are "SFIM" (Smoothing
        Filter-based Intensity Modulation), "HPFA" (High Pass Filter Addition), "PCS"
        (Principal Component Substitution), and "SM" (simple mean). Different sharpening
        methods will produce different quality sharpening results in different scenarios.
    qa : str | list, default=None
        One or more optional quality assessment names to apply after sharpening. Results 
        will be stored as image properties with the pattern `eemont:metric`, e.g. `eemont:RMSE`.
    **kwargs :
        Keyword arguments passed to ee.Image.reduceRegion() such as "geometry",
        "maxPixels", "bestEffort", etc. These arguments are only used for PCS sharpening
        and quality assessments.

    Returns
    -------
    ee.Image
        The Image with all sharpenable bands sharpened to the panchromatic resolution and
        quality assessments run and set as properties.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> source = ee.Image("LANDSAT/LC08/C01/T1_TOA/LC08_047027_20160819")
    >>> sharp = source.panSharpen(method="HPFA", qa=["MSE", "RMSE"], maxPixels=1e13)
    """
    return ee_extra.Algorithms.core.panSharpen(img=self, method=method, qa=qa, prefix="eemont", **kwargs)


@extend(ee.image.Image)
def matchHistogram(self, target, bands, geometry=None, maxBuckets=256):
    """Adjust the image's histogram to match a target image.

    Parameters
    ----------
    self : ee.Image [this]
        Image to adjust.
    target : ee.Image
        Image to match.
    bands : dict
        A dictionary of band names to match, with source bands as keys and target bands as
        values.
    geometry : ee.Geometry, default=None
        The region to match histograms in that overlaps both images. If none is provided,
        the geometry of the source image will be used.
    maxBuckets : int, default=256
        The maximum number of buckets to use when building histograms. Will be rounded to
        the nearest power of 2.

    Returns
    -------
    ee.Image
        The adjusted image containing the matched source bands.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> source = ee.Image("LANDSAT/LC08/C01/T1_TOA/LC08_047027_20160819")
    >>> target = ee.Image("LANDSAT/LE07/C01/T1_TOA/LE07_046027_20150701")
    >>> bands = {
    >>>    "B4": "B3",
    >>>    "B3": "B2",
    >>>    "B2": "B1"
    >>> }
    >>> matched = source.matchHistogram(target, bands)
    """
    return ee_extra.Spectral.core.matchHistogram(
        source=self, 
        target=target, 
        bands=bands, 
        geometry=geometry, 
        maxBuckets=maxBuckets
    )


@extend(ee.image.Image)
def tasseledCap(self):
    """Calculates tasseled cap brightness, wetness, and greenness components.

    Tasseled cap transformations are applied using coefficients published for these
    supported platforms:

    * Sentinel-2 MSI Level 1C [1]_
    * Landsat 9 OLI-2 SR [2]_
    * Landsat 9 OLI-2 TOA [2]_
    * Landsat 8 OLI SR [2]_
    * Landsat 8 OLI TOA [2]_
    * Landsat 7 ETM+ TOA [3]_
    * Landsat 5 TM Raw DN [4]_
    * Landsat 4 TM Raw DN [5]_
    * Landsat 4 TM Surface Reflectance [6]_
    * MODIS NBAR [7]_

    Parameters
    ----------
    self : ee.Image
        Image to calculate tasseled cap components for. Must belong to a supported
        platform.

    Returns
    -------
    ee.Image
        Image with the tasseled cap components as new bands.

    References
    ----------
    .. [1] Shi, T., & Xu, H. (2019). Derivation of Tasseled Cap Transformation
        Coefficients for Sentinel-2 MSI At-Sensor Reflectance Data. IEEE Journal
        of Selected Topics in Applied Earth Observations and Remote Sensing, 1–11.
        doi:10.1109/jstars.2019.2938388
    .. [2] Zhai, Y., Roy, D.P., Martins, V.S., Zhang, H.K., Yan, L., Li, Z. 2022.
        Conterminous United States Landsat-8 top of atmosphere and surface reflectance
        tasseled cap transformation coefficeints. Remote Sensing of Environment, 
        274(2022). doi:10.1016/j.rse.2022.112992
    .. [3] Huang, C., Wylie, B., Yang, L., Homer, C. and Zylstra, G., 2002.
        Derivation of a tasselled cap transformation based on Landsat 7 at-satellite
        reflectance. International journal of remote sensing, 23(8), pp.1741-1748.
    .. [4] Crist, E.P., Laurin, R. and Cicone, R.C., 1986, September. Vegetation and
        soils information contained in transformed Thematic Mapper data. In
        Proceedings of IGARSS’86 symposium (pp. 1465-1470). Paris: European Space
        Agency Publications Division.
    .. [5] Crist, E.P. and Cicone, R.C., 1984. A physically-based transformation of
        Thematic Mapper data---The TM Tasseled Cap. IEEE Transactions on Geoscience
        and Remote sensing, (3), pp.256-263.
    .. [6] Crist, E.P., 1985. A TM tasseled cap equivalent transformation for
        reflectance factor data. Remote sensing of Environment, 17(3), pp.301-306.
    .. [7] Lobser, S.E. and Cohen, W.B., 2007. MODIS tasselled cap: land cover
        characteristics expressed through transformed MODIS data. International
        Journal of Remote Sensing, 28(22), pp.5079-5101.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> img = ee.Image("LANDSAT/LT05/C01/T1/LT05_044034_20081011")
    >>> img = img.tasseledCap()
    """
    return ee_extra.Spectral.core.tasseledCap(self)



import ee
import geopy
from geopy.geocoders import get_geocoder_for_service

from .common import _retrieve_location
from .extending import extend
from .geometry import *


@extend(ee.featurecollection.FeatureCollection)
def __len__(self):
    """Returns the size of the feature collection.

    Parameters
    ----------
    self : ee.FeatureCollection
        Feature Collection to get the size from.

    Returns
    -------
    int
        Size of the feature collection.
    """
    return self.size().getInfo()


@extend(ee.featurecollection.FeatureCollection)
def __getitem__(self, key):
    """Gets the column of each feature in the feature collection according to the
    specified key.

    Parameters
    ----------
    self : ee.FeatureCollection
        Feature Collection to get the columns from.
    key : string | list[string]
        Key used to get the specified column. If string, it gets the column with that name
        or that matches with regex. If list, it gets multiple columns.

    Returns
    -------
    ee.FeatureCollection
        Feature Collection with the selected columns.
    """
    return self.select(key)


@extend(ee.featurecollection.FeatureCollection, static=True)
def MultiPointFromQuery(query, geocoder="nominatim", **kwargs):
    """Constructs an ee.Feature describing a point from a query submitted to a geodocer
    using the geopy package.

    This returns all pairs of coordinates retrieved by the query. The properties of the
    feature collection correspond to the raw properties retrieved by the locations of
    the query.

    Tip
    ----------
    Check more info about constructors in the :ref:`User Guide<Constructors>`.

    Parameters
    ----------
    query : str
        Address, query or structured query to geocode.
    geocoder : str, default = 'nominatim'
        Geocoder to use. Please visit https://geopy.readthedocs.io/ for more info.
    **kwargs :
        Keywords arguments for geolocator.geocode(). The user_agent argument is mandatory
        (this argument can be set as user_agent = 'my-gee-username' or user_agent =
        'my-gee-app-name'). Please visit https://geopy.readthedocs.io/ for more info.

    Returns
    -------
    ee.FeatureCollection
        Feature Collection with point geometries from the specified query.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.FeatureCollection.MultiPointFromQuery('Río Amazonas',
    ...                                     user_agent = 'my-gee-eemont-query').getInfo()
    {'type': 'FeatureCollection',
     'columns': {'boundingbox': 'List<String>',
      'class': 'String',
      'display_name': 'String',
      'importance': 'Float',
      'lat': 'String',
      'licence': 'String',
      'lon': 'String',
      'osm_id': 'Integer',
      'osm_type': 'String',
      'place_id': 'Integer',
      'system:index': 'String',
      'type': 'String'},
     'features': [{'type': 'Feature',
       'geometry': {'type': 'Point', 'coordinates': [-57.4801276, -2.3740229]},
       'id': '0',
       'properties': {'boundingbox': ['-4.4421898',
         '0.7065296',
         '-73.4501259',
         '-49.2759133'],
        'class': 'waterway',
        'display_name': 'Rio Amazonas, Região Norte, 69100-143, Brasil',
        'importance': 0.4,
        'lat': '-2.3740229',
        'licence': 'Data © OpenStreetMap contributors, ODbL 1.0. https://osm.org/copyright',
        'lon': '-57.4801276',
        'osm_id': 2295651,
        'osm_type': 'relation',
        'place_id': 258650987,
        'type': 'river'}},
      {'type': 'Feature',
       'geometry': {'type': 'Point',
        'coordinates': [-70.04978704421745, -4.10958645]},
       'id': '1',
       'properties': {'boundingbox': ['-4.2647706',
         '-3.9548576',
         '-70.1817875',
         '-69.9440055'],
        'class': 'natural',
        'display_name': 'Río Amazonas, Ramón Castilla, Mariscal Ramón Castilla, Loreto, Perú',
        'importance': 0.39999999999999997,
        'lat': '-4.10958645',
        'licence': 'Data © OpenStreetMap contributors, ODbL 1.0. https://osm.org/copyright',
        'lon': '-70.04978704421745',
        'osm_id': 8495385,
        'osm_type': 'relation',
        'place_id': 297654614,
        'type': 'water'}},
      ...]}
    """
    locations = _retrieve_location(query, geocoder, False, **kwargs)

    features = []

    for location in locations:
        geometry = ee.Geometry.Point([location.longitude, location.latitude])
        feature = ee.Feature(geometry, location.raw)
        features.append(feature)

    return ee.FeatureCollection(features)



import warnings

import ee

from .extending import extend


@extend(ee.dictionary.Dictionary)
def __contains__(self, key):
    """Returns True if the key is in the dictionary.

    Parameters
    ----------
    self : ee.Dictionary
        Dictionary to check.
    key : item
        Item to check.

    Returns
    -------
    boolean
        Whether the key is contained in the dictionary.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> eeDict = ee.Dictionary({"a":1,"b":2,"h":4})
    >>> "a" in eeDict
    True
    >>> "c" in eeDict
    False
    """
    return self.contains(key).getInfo()


@extend(ee.dictionary.Dictionary)
def __getitem__(self, key):
    """Gets the item of the dictionary according to the specified key.

    Parameters
    ----------
    self : ee.Dictionary
        Dictionary to get the items from.
    key : str
        Key used to get the specified item. It gets the value of the specified key.

    Returns
    -------
    ee.Element
        Selected value.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> eeDict = ee.Dictionary({"a":1,"b":2,"h":4})
    >>> eeDict["a"].getInfo()
    1
    >>> eeDict["h"].getInfo()
    4
    """
    return self.get(key)



import ee
import geopy
from geopy.geocoders import get_geocoder_for_service

from .common import (_convert_lnglats_to_pluscodes,
                     _convert_pluscodes_to_lnglats, _lnglat_from_location,
                     _retrieve_location)
from .extending import extend


@extend(ee.geometry.Geometry, static=True)
def BBoxFromQuery(query, geocoder="nominatim", **kwargs):
    """Constructs an ee.Geometry describing a bounding box from a query submitted to a
    geodocer using the geopy package.

    Tip
    ----------
    Check more info about constructors in the :ref:`User Guide<Constructors>`.

    Parameters
    ----------
    query : str
        Address, query or structured query to geocode.
    geocoder : str, default = 'nominatim'
        Geocoder to use. One of 'nominatim' or 'arcgis'. Please visit
        https://geopy.readthedocs.io/ for more info.
    **kwargs :
        Keywords arguments for geolocator.geocode(). The user_agent argument is mandatory
        (this argument can be set as user_agent = 'my-gee-username' or user_agent =
        'my-gee-app-name'). Please visit https://geopy.readthedocs.io/ for more info.

    Returns
    -------
    ee.Geometry.Polygon
        Geometry describing a bounding box from the specified query.

    See Also
    --------
    PointFromQuery : Constructs an ee.Geometry describing a point from a query submitted
        to a geodocer using the geopy package.
    MultiPointFromQuery : Constructs an ee.Geometry describing a multi-point from a query
        submitted to a geodocer using the geopy package.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.Geometry.BBoxFromQuery('Seattle',user_agent = 'my-gee-eemont-query')
    ee.Geometry({
      "functionInvocationValue": {
        "functionName": "GeometryConstructors.Polygon",
        "arguments": {
          "coordinates": {
            "constantValue": [
              [
                [
                  -122.45969600000001,
                  47.7341357
                ],
                [
                  -122.45969600000001,
                  47.4810022
                ],
                [
                  -122.224433,
                  47.4810022
                ],
                [
                  -122.224433,
                  47.7341357
                ]
              ]
            ]
          },
          "geodesic": {
            "constantValue": false
          }
        }
      }
    })
    """
    if geocoder not in ["nominatim", "arcgis"]:
        raise Exception('Invalid geocoder! Use one of "nominatim" or "arcgis".')

    location = _retrieve_location(query, geocoder, True, **kwargs)

    if geocoder == "nominatim":
        BBox = location.raw["boundingbox"]
        return ee.Geometry.BBox(
            float(BBox[2]), float(BBox[0]), float(BBox[3]), float(BBox[1])
        )
    elif geocoder == "arcgis":
        BBox = location.raw["extent"]
        return ee.Geometry.BBox(BBox["xmin"], BBox["ymin"], BBox["xmax"], BBox["ymax"])
    else:
        raise Exception('Invalid geocoder! Use one of "nominatim" or "arcgis".')


@extend(ee.geometry.Geometry, static=True)
def PointFromQuery(query, geocoder="nominatim", **kwargs):
    """Constructs an ee.Geometry describing a point from a query submitted to a geodocer
    using the geopy package.

    This returns exactly one pair of coordinates.

    Tip
    ----------
    Check more info about constructors in the :ref:`User Guide<Constructors>`.

    Parameters
    ----------
    query : str
        Address, query or structured query to geocode.
    geocoder : str, default = 'nominatim'
        Geocoder to use. Please visit https://geopy.readthedocs.io/ for more info.
    **kwargs :
        Keywords arguments for geolocator.geocode(). The user_agent argument is mandatory
        (this argument can be set as user_agent = 'my-gee-username' or user_agent =
        'my-gee-app-name'). Please visit https://geopy.readthedocs.io/ for more info.

    Returns
    -------
    ee.Geometry.Point
        Geometry describing a point from the specified query.

    See Also
    --------
    BBoxFromQuery : Constructs an ee.Geometry describing a bounding box from a query
        submitted to a geodocer using the geopy package.
    MultiPointFromQuery : Constructs an ee.Geometry describing a multi-point from a query
        submitted to a geodocer using the geopy package.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.Geometry.PointFromQuery('Mt. Rainier, USA',user_agent = 'my-gee-eemont-query')
    ee.Geometry({
      "functionInvocationValue": {
        "functionName": "GeometryConstructors.Point",
        "arguments": {
          "coordinates": {
            "constantValue": [
              -121.757682,
              46.8521484
            ]
          }
        }
      }
    })
    """
    location = _retrieve_location(query, geocoder, True, **kwargs)
    return ee.Geometry.Point(_lnglat_from_location(location))


@extend(ee.geometry.Geometry, static=True)
def MultiPointFromQuery(query, geocoder="nominatim", **kwargs):
    """Constructs an ee.Geometry describing a multi-point from a query submitted to a
    geodocer using the geopy package.

    This returns all pairs of coordinates retrieved by the query.

    Tip
    ----------
    Check more info about constructors in the :ref:`User Guide<Constructors>`.

    Parameters
    ----------
    query : str
        Address, query or structured query to geocode.
    geocoder : str, default = 'nominatim'
        Geocoder to use. Please visit https://geopy.readthedocs.io/ for more info.
    **kwargs :
        Keywords arguments for geolocator.geocode(). The user_agent argument is mandatory
        (this argument can be set as user_agent = 'my-gee-username' or user_agent =
        'my-gee-app-name'). Please visit https://geopy.readthedocs.io/ for more info.

    Returns
    -------
    ee.Geometry.MultiPoint
        Geometry describing a multi-point from the specified query.

    See Also
    --------
    BBoxFromQuery : Constructs an ee.Geometry describing a bounding box from a query
        submitted to a geodocer using the geopy package.
    PointFromQuery : Constructs an ee.Geometry describing a point from a query submitted
        to a geodocer using the geopy package.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.Geometry.MultiPointFromQuery('Mt. Rainier, USA',user_agent = 'my-gee-eemont-query')
    ee.Geometry({
      "functionInvocationValue": {
        "functionName": "GeometryConstructors.MultiPoint",
        "arguments": {
          "coordinates": {
            "constantValue": [
              [
                -121.757682,
                46.8521484
              ],
              [
                -76.9649751,
                38.9415
              ],
              [
                -97.78689,
                30.154197
              ],
              [
                -84.130098,
                39.863452
              ]
            ]
          }
        }
      }
    })
    """
    locations = _retrieve_location(query, geocoder, False, **kwargs)
    coords = []
    for location in locations:
        coords.append([location.longitude, location.latitude])
    return ee.Geometry.MultiPoint(coords)


@extend(ee.geometry.Geometry, static=True)
def PointFromPlusCode(pluscode, geocoder="nominatim", **kwargs):
    """Constructs an ee.Geometry describing a point from a Plus Code.

    If the Plus Code is full, it will be decoded directly using the openlocationcode
    package. If it is a short Code with a reference location, the reference will be
    geocoded using the geopy package.

    Tip
    ----------
    Check more info about constructors in the :ref:`User Guide<Constructors>`.

    Parameters
    ----------
    pluscode : str
        A full Plus Code or a short Plus Code with a geocodable reference location appended to it.
    geocoder : str, default = 'nominatim'
        Geocoder to use. Please visit https://geopy.readthedocs.io/ for more info.
    **kwargs :
        Keywords arguments for geolocator.geocode(). The user_agent argument is mandatory
        (this argument can be set as user_agent = 'my-gee-username' or user_agent =
        'my-gee-app-name'). Please visit https://geopy.readthedocs.io/ for more info.

    Returns
    -------
    ee.Geometry.Point
        Geometry describing a point from the Plus Code centroid.

    See Also
    --------
    MultiPointFromPlusCodes : Constructs an ee.Geometry describing multiple points from a
        list of Plus Codes.
    PolygonFromPlusCodes : Constructs an ee.Geometry describing a polygon from a list of
        Plus Codes.
    MultiPolygonFromPlusCodes : Constructs an ee.Geometry describing multiple polygons
        from a list of lists of Plus Codes.
    LineStringFromPlusCodes : Constructs an ee.Geometry describing a line from a list of
        Plus Codes.
    MultiLineStringFromPlusCodes : Constructs an ee.Geometry describing multiple lines
        from a list of lists of Plus Codes.
    LinearRingFromPlusCodes : Constructs an ee.Geometry describing a linear ring from a
        list of Plus Codes.
    RectangleFromPlusCodes : Constructs an ee.Geometry describing a rectangle from a list
        of two Plus Code corners.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> geom = ee.Geometry.PointFromPlusCode('QXGV+XH Denver, CO, USA',
    ...                                      user_agent = 'my-gee-eemont-query')
    >>> geom.coordinates().getInfo()
    [-105.0060625, 39.777437500000005]
    """
    coordinates = _convert_pluscodes_to_lnglats(pluscode, geocoder, **kwargs)
    return ee.Geometry.Point(coordinates)


@extend(ee.geometry.Geometry, static=True)
def MultiPointFromPlusCodes(pluscodes, geocoder="nominatim", **kwargs):
    """Constructs an ee.Geometry describing multiple points from a list of Plus Codes.

    If the Plus Codes are full, they will be decoded directly using the openlocationcode
    package. If they are short Codes with reference locations, the references will be
    geocoded using the geopy package.

    Tip
    ----------
    Check more info about constructors in the :ref:`User Guide<Constructors>`.

    Parameters
    ----------
    pluscodes : str
        A list of full Plus Codes or short Plus Codes with geocodable reference locations
        appended to them.
    geocoder : str, default = 'nominatim'
        Geocoder to use. Please visit https://geopy.readthedocs.io/ for more info.
    **kwargs :
        Keywords arguments for geolocator.geocode(). The user_agent argument is mandatory
        (this argument can be set as user_agent = 'my-gee-username' or user_agent =
        'my-gee-app-name'). Please visit https://geopy.readthedocs.io/ for more info.

    Returns
    -------
    ee.Geometry.MultiPoint
        Geometry describing multiple points from the Plus Code centroids.

    See Also
    --------
    PointFromPlusCode : Constructs an ee.Geometry describing a point from a Plus Code.
    PolygonFromPlusCodes : Constructs an ee.Geometry describing a polygon from a list of
        Plus Codes.
    MultiPolygonFromPlusCodes : Constructs an ee.Geometry describing multiple polygons
        from a list of lists of Plus Codes.
    LineStringFromPlusCodes : Constructs an ee.Geometry describing a line from a list of
        Plus Codes.
    MultiLineStringFromPlusCodes : Constructs an ee.Geometry describing multiple lines
        from a list of lists of Plus Codes.
    LinearRingFromPlusCodes : Constructs an ee.Geometry describing a linear ring from a
        list of Plus Codes.
    RectangleFromPlusCodes : Constructs an ee.Geometry describing a rectangle from a list
        of two Plus Code corners.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> geom = ee.Geometry.MultiPointFromPlusCodes(['85FQ2222+22','85FR2222+22'])
    >>> geom.coordinates().getInfo()
    [[-104.9999375, 39.0000625], [-103.9999375, 39.0000625]]
    """
    coordinates = _convert_pluscodes_to_lnglats(pluscodes, geocoder, **kwargs)
    return ee.Geometry.MultiPoint(coordinates)


@extend(ee.geometry.Geometry, static=True)
def PolygonFromPlusCodes(pluscodes, geocoder="nominatim", **kwargs):
    """Constructs an ee.Geometry describing a polygon from a list of Plus Codes.

    If the Plus Codes are full, they will be decoded directly using the openlocationcode
    package. If they are short Codes with reference locations, the references will be
    geocoded using the geopy package.

    Tip
    ----------
    Check more info about constructors in the :ref:`User Guide<Constructors>`.

    Parameters
    ----------
    pluscodes : str
        A list of full Plus Codes or short Plus Codes with geocodable reference locations
        appended to them.
    geocoder : str, default = 'nominatim'
        Geocoder to use. Please visit https://geopy.readthedocs.io/ for more info.
    **kwargs :
        Keywords arguments for geolocator.geocode(). The user_agent argument is mandatory
        (this argument can be set as user_agent = 'my-gee-username' or user_agent =
        'my-gee-app-name'). Please visit https://geopy.readthedocs.io/ for more info.

    Returns
    -------
    ee.Geometry.Polygon
        Geometry describing a polygon from the Plus Code centroids.

    See Also
    --------
    PointFromPlusCode : Constructs an ee.Geometry describing a point from a Plus Code.
    MultiPointFromPlusCodes : Constructs an ee.Geometry describing multiple points from a
        list of Plus Codes.
    MultiPolygonFromPlusCodes : Constructs an ee.Geometry describing multiple polygons
        from a list of lists of Plus Codes.
    LineStringFromPlusCodes : Constructs an ee.Geometry describing a line from a list of
        Plus Codes.
    MultiLineStringFromPlusCodes : Constructs an ee.Geometry describing multiple lines
        from a list of lists of Plus Codes.
    LinearRingFromPlusCodes : Constructs an ee.Geometry describing a linear ring from a
        list of Plus Codes.
    RectangleFromPlusCodes : Constructs an ee.Geometry describing a rectangle from a list
        of two Plus Code corners.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> geom = ee.Geometry.PolygonFromPlusCodes(['85FQ2222+22','85FR2222+22','85GR2222+22'])
    >>> geom.coordinates().getInfo()
    [[[-104.9999375, 39.0000625],
      [-103.9999375, 39.0000625],
      [-103.9999375, 40.0000625],
      [-104.9999375, 39.0000625]]]
    """
    coordinates = _convert_pluscodes_to_lnglats(pluscodes, geocoder, **kwargs)
    return ee.Geometry.Polygon(coordinates)


@extend(ee.geometry.Geometry, static=True)
def MultiPolygonFromPlusCodes(pluscodes, geocoder="nominatim", **kwargs):
    """Constructs an ee.Geometry describing multiple polygons from a list of lists of
    Plus Codes.

    If the Plus Codes are full, they will be decoded directly using the openlocationcode
    package. If they are short Codes with reference locations, the references will be
    geocoded using the geopy package.

    Tip
    ----------
    Check more info about constructors in the :ref:`User Guide<Constructors>`.

    Parameters
    ----------
    pluscodes : str
        A list of lists of full Plus Codes or short Plus Codes with geocodable reference
        locations appended to them.
    geocoder : str, default = 'nominatim'
        Geocoder to use. Please visit https://geopy.readthedocs.io/ for more info.
    **kwargs :
        Keywords arguments for geolocator.geocode(). The user_agent argument is mandatory
        (this argument can be set as user_agent = 'my-gee-username' or user_agent =
        'my-gee-app-name'). Please visit https://geopy.readthedocs.io/ for more info.

    Returns
    -------
    ee.Geometry.MultiPolygon
        Geometry describing multiple polygons from the Plus Code centroids.

    See Also
    --------
    PointFromPlusCode : Constructs an ee.Geometry describing a point from a Plus Code.
    MultiPointFromPlusCodes : Constructs an ee.Geometry describing multiple points from a
        list of Plus Codes.
    PolygonFromPlusCodes : Constructs an ee.Geometry describing a polygon from a list of
        Plus Codes.
    LineStringFromPlusCodes : Constructs an ee.Geometry describing a line from a list of
        Plus Codes.
    MultiLineStringFromPlusCodes : Constructs an ee.Geometry describing multiple lines
        from a list of lists of Plus Codes.
    LinearRingFromPlusCodes : Constructs an ee.Geometry describing a linear ring from a
        list of Plus Codes.
    RectangleFromPlusCodes : Constructs an ee.Geometry describing a rectangle from a list
        of two Plus Code corners.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> geom = ee.Geometry.MultiPolygonFromPlusCodes([
    ...    ['85FQ2222+22', '85FR2222+22', '85GR2222+22'],
    ...    ['85FP8PC2+G2', '85FPJF23+G4', '85FPMW2R+RP']
    ...  ])
    >>> geom.coordinates().getInfo()
    [[[[-104.9999375, 39.0000625],
      [-103.9999375, 39.0000625],
      [-103.9999375, 40.0000625],
      [-104.9999375, 39.0000625]],
      [[-105.2999375, 39.321312500000005],
      [-105.5471875, 39.6013125],
      [-105.0581875, 39.6520625],
      [-105.2999375, 39.321312500000005]]]]
    """
    coordinates = _convert_pluscodes_to_lnglats(pluscodes, geocoder, **kwargs)
    return ee.Geometry.MultiPolygon(coordinates)


@extend(ee.geometry.Geometry, static=True)
def LineStringFromPlusCodes(pluscodes, geocoder="nominatim", **kwargs):
    """Constructs an ee.Geometry describing a line from a list of Plus Codes.

    If the Plus Codes are full, they will be decoded directly using the openlocationcode
    package. If they are short Codes with reference locations, the references will be
    geocoded using the geopy package.

    Tip
    ----------
    Check more info about constructors in the :ref:`User Guide<Constructors>`.

    Parameters
    ----------
    pluscodes : str
        A list of full Plus Codes or short Plus Codes with geocodable reference locations
        appended to them.
    geocoder : str, default = 'nominatim'
        Geocoder to use. Please visit https://geopy.readthedocs.io/ for more info.
    **kwargs :
        Keywords arguments for geolocator.geocode(). The user_agent argument is mandatory
        (this argument can be set as user_agent = 'my-gee-username' or user_agent =
        'my-gee-app-name'). Please visit https://geopy.readthedocs.io/ for more info.

    Returns
    -------
    ee.Geometry.LineString
        Geometry describing a line from the Plus Code centroids.

    See Also
    --------
    PointFromPlusCode : Constructs an ee.Geometry describing a point from a Plus Code.
    MultiPointFromPlusCodes : Constructs an ee.Geometry describing multiple points from a
        list of Plus Codes.
    PolygonFromPlusCodes : Constructs an ee.Geometry describing a polygon from a list of
        Plus Codes.
    MultiPolygonFromPlusCodes : Constructs an ee.Geometry describing multiple polygons
        from a list of lists of Plus Codes.
    MultiLineStringFromPlusCodes : Constructs an ee.Geometry describing multiple lines
        from a list of lists of Plus Codes.
    LinearRingFromPlusCodes : Constructs an ee.Geometry describing a linear ring from a
        list of Plus Codes.
    RectangleFromPlusCodes : Constructs an ee.Geometry describing a rectangle from a list
        of two Plus Code corners.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> geom = ee.Geometry.LineStringFromPlusCodes(['85FQ2222+22','85FR2222+22'])
    >>> geom.coordinates().getInfo()
    [[-104.9999375, 39.0000625], [-103.9999375, 39.0000625]]
    """
    coordinates = _convert_pluscodes_to_lnglats(pluscodes, geocoder, **kwargs)
    return ee.Geometry.LineString(coordinates)


@extend(ee.geometry.Geometry, static=True)
def MultiLineStringFromPlusCodes(pluscodes, geocoder="nominatim", **kwargs):
    """Constructs an ee.Geometry describing multiple lines from a list of lists of Plus
    Codes.

    If the Plus Codes are full, they will be decoded directly using the openlocationcode
    package. If they are short Codes with reference locations, the references will be
    geocoded using the geopy package.

    Tip
    ----------
    Check more info about constructors in the :ref:`User Guide<Constructors>`.

    Parameters
    ----------
    pluscodes : str
        A list of lists of full Plus Codes or short Plus Codes with geocodable reference
        locations appended to them.
    geocoder : str, default = 'nominatim'
        Geocoder to use. Please visit https://geopy.readthedocs.io/ for more info.
    **kwargs :
        Keywords arguments for geolocator.geocode(). The user_agent argument is mandatory
        (this argument can be set as user_agent = 'my-gee-username' or user_agent =
        'my-gee-app-name'). Please visit https://geopy.readthedocs.io/ for more info.

    Returns
    -------
    ee.Geometry.MultiLineString
        Geometry describing multiple lines from the Plus Code centroids.

    See Also
    --------
    PointFromPlusCode : Constructs an ee.Geometry describing a point from a Plus Code.
    MultiPointFromPlusCodes : Constructs an ee.Geometry describing multiple points from a
        list of Plus Codes.
    PolygonFromPlusCodes : Constructs an ee.Geometry describing a polygon from a list of
        Plus Codes.
    MultiPolygonFromPlusCodes : Constructs an ee.Geometry describing multiple polygons
        from a list of lists of Plus Codes.
    LineStringFromPlusCodes : Constructs an ee.Geometry describing a line from a list of
        Plus Codes.
    LinearRingFromPlusCodes : Constructs an ee.Geometry describing a linear ring from a
        list of Plus Codes.
    RectangleFromPlusCodes : Constructs an ee.Geometry describing a rectangle from a list
        of two Plus Code corners.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> geom = ee.Geometry.MultiLineStringFromPlusCodes([
    ...    ['85FQ2222+22', '85FR2222+22'],
    ...    ['85FP8PC2+G2', '85FPJF23+G4']
    ... ])
    >>> geom.coordinates().getInfo()
    [[[-104.9999375, 39.0000625], [-103.9999375, 39.0000625]],
    [[-105.2999375, 39.321312500000005], [-105.5471875, 39.6013125]]]
    """
    coordinates = _convert_pluscodes_to_lnglats(pluscodes, geocoder, **kwargs)
    return ee.Geometry.MultiLineString(coordinates)


@extend(ee.geometry.Geometry, static=True)
def LinearRingFromPlusCodes(pluscodes, geocoder="nominatim", **kwargs):
    """Constructs an ee.Geometry describing a linear ring from a list of Plus Codes.

    If the Plus Codes are full, they will be decoded directly using the openlocationcode
    package. If they are short Codes with reference locations, the references will be
    geocoded using the geopy package.

    Tip
    ----------
    Check more info about constructors in the :ref:`User Guide<Constructors>`.

    Parameters
    ----------
    pluscodes : str
        A list of full Plus Codes or short Plus Codes with geocodable reference locations
        appended to them.
    geocoder : str, default = 'nominatim'
        Geocoder to use. Please visit https://geopy.readthedocs.io/ for more info.
    **kwargs :
        Keywords arguments for geolocator.geocode(). The user_agent argument is mandatory
        (this argument can be set as user_agent = 'my-gee-username' or user_agent =
        'my-gee-app-name'). Please visit https://geopy.readthedocs.io/ for more info.

    Returns
    -------
    ee.Geometry.LinearRing
        Geometry describing a linear ring from the Plus Code centroids.

    See Also
    --------
    PointFromPlusCode : Constructs an ee.Geometry describing a point from a Plus Code.
    MultiPointFromPlusCodes : Constructs an ee.Geometry describing multiple points from a
        list of Plus Codes.
    PolygonFromPlusCodes : Constructs an ee.Geometry describing a polygon from a list of
        Plus Codes.
    MultiPolygonFromPlusCodes : Constructs an ee.Geometry describing multiple polygons
        from a list of lists of Plus Codes.
    LineStringFromPlusCodes : Constructs an ee.Geometry describing a line from a list of
        Plus Codes.
    MultiLineStringFromPlusCodes : Constructs an ee.Geometry describing multiple lines
        from a list of lists of Plus Codes.
    RectangleFromPlusCodes : Constructs an ee.Geometry describing a rectangle from a list
        of two Plus Code corners.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> geom = ee.Geometry.LinearRingFromPlusCodes(['85FQ2222+22',
    ...                                             '85FR2222+22', '85GR2222+22'])
    >>> geom.coordinates().getInfo()
    [[-104.9999375, 39.0000625],
    [-103.9999375, 39.0000625],
    [-103.9999375, 40.0000625],
    [-104.9999375, 39.0000625]]
    """
    coordinates = _convert_pluscodes_to_lnglats(pluscodes, geocoder, **kwargs)
    return ee.Geometry.LinearRing(coordinates)


@extend(ee.geometry.Geometry, static=True)
def RectangleFromPlusCodes(pluscodes, geocoder="nominatim", **kwargs):
    """Constructs an ee.Geometry describing a rectangle from a list of two Plus Code
    corners.

    If the Plus Codes are full, they will be decoded directly using the openlocationcode
    package. If they are short Codes with reference locations, the references will be
    geocoded using the geopy package.

    Tip
    ----------
    Check more info about constructors in the :ref:`User Guide<Constructors>`.

    Parameters
    ----------
    pluscodes : str
        A list of two full Plus Codes or short Plus Codes with geocodable reference
        locations appended to them describing the corners of the rectangle.
    geocoder : str, default = 'nominatim'
        Geocoder to use. Please visit https://geopy.readthedocs.io/ for more info.
    **kwargs :
        Keywords arguments for geolocator.geocode(). The user_agent argument is mandatory
        (this argument can be set as user_agent = 'my-gee-username' or user_agent =
        'my-gee-app-name'). Please visit https://geopy.readthedocs.io/ for more info.

    Returns
    -------
    ee.Geometry.Rectangle
        Geometry describing a rectangle from the Plus Code centroids.

    See Also
    --------
    PointFromPlusCode : Constructs an ee.Geometry describing a point from a Plus Code.
    MultiPointFromPlusCodes : Constructs an ee.Geometry describing multiple points from a
        list of Plus Codes.
    PolygonFromPlusCodes : Constructs an ee.Geometry describing a polygon from a list of
        Plus Codes.
    MultiPolygonFromPlusCodes : Constructs an ee.Geometry describing multiple polygons
        from a list of lists of Plus Codes.
    LineStringFromPlusCodes : Constructs an ee.Geometry describing a line from a list of
        Plus Codes.
    MultiLineStringFromPlusCodes : Constructs an ee.Geometry describing multiple lines
        from a list of lists of Plus Codes.
    LinearRingFromPlusCodes : Constructs an ee.Geometry describing a linear ring from a
        list of Plus Codes.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> geom = ee.Geometry.RectangleFromPlusCodes(['85FQ2222+22','85FR2222+22'])
    >>> geom.coordinates().getInfo()
    [[[-104.9999375, 39.0000625],
      [-104.9999375, 39.0000625],
      [-103.9999375, 39.0000625],
      [-103.9999375, 39.0000625],
      [-104.9999375, 39.0000625]]]
    """
    coordinates = _convert_pluscodes_to_lnglats(pluscodes, geocoder, **kwargs)
    return ee.Geometry.Rectangle(coordinates)


@extend(ee.geometry.Geometry, static=False)
def plusCodes(self, codeLength=10):
    """Converts the coordinates of an ee.Geometry to Plus Codes.

    Parameters
    ----------
    self : ee.Geometry
        The geometry to extract coordinates from.
    codeLength : int, default = 10
        The number of significant digits in the output codes, between 2 and 15. Shorter
        codes are less precise.

    Returns
    -------
    list | str
        The coordinates of the geometry converted to Plus Codes. The structure of the Plus
        Codes array will be identical to the structure returned by
        ee.Geometry.coordinates().

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> pt = ee.Geometry.Point([-105, 40])
    >>> pt.plusCodes()
    '85GQ2222+22'
    """
    coordinates = self.coordinates().getInfo()
    plus_codes = _convert_lnglats_to_pluscodes(coordinates, codeLength)
    return plus_codes



import warnings

import ee

from .extending import extend


@extend(ee.ee_list.List)
def __contains__(self, key):
    """Returns True if the item is in the list.

    Parameters
    ----------
    self : ee.List
        List to check.
    key : item
        Item to check.

    Returns
    -------
    boolean
        Whether the item is contained in the list.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> eeList = ee.List([1,2,3,5])
    >>> 1 in eeList
    True
    >>> 3 in eeList
    False
    """
    return self.contains(key).getInfo()


@extend(ee.ee_list.List)
def __len__(self):
    """Returns the length of the list.

    Parameters
    ----------
    self : ee.List
        List to get the length from.

    Returns
    -------
    int
        Length of the list.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> eeList = ee.List([1,2,3,5])
    >>> len(eeList)
    4
    """
    return self.length().getInfo()


@extend(ee.ee_list.List)
def __getitem__(self, key):
    """Gets the item of the list according to the specified key.

    Parameters
    ----------
    self : ee.List
        List to get the items from.
    key : numeric | list[numeric] | slice
        Key used to get the specified item. If numeric, it gets the item at that index.
        If list, it gets multiple items. If slice, it calls the slice() method (the step
        parameter is ignored).

    Returns
    -------
    ee.List | ee.Number
        List or number with the selected items.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> eeList = ee.List([1,2,3,5])
    >>> eeList[0].getInfo()
    1
    >>> eeList[0:2].getInfo()
    [1,2,3]
    >>> eeList[0:].getInfo()
    [1,2,3,5]
    >>> eeList[[0,2]].getInfo()
    [1,3]
    """
    if isinstance(key, slice):

        if key.start == None:
            start = 0
        else:
            start = key.start

        if key.stop == None:
            stop = self.length()
        else:
            stop = key.stop

        selected = self.slice(start, stop)

    elif isinstance(key, list):

        selected = ee.List([])

        for k in key:
            selected = selected.add(self.get(k))

    else:
        selected = self.get(key)

    return selected


@extend(ee.ee_list.List)
def __add__(self, other):
    """Concatenates the contents of a list onto another list.

    Parameters
    ----------
    self : ee.List
        Left operand.
    other : ee.List | list
        Right operand.

    Returns
    -------
    ee.List
        Concatenation of two lists.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> eeList1 = ee.List([1,2,3,4])
    >>> eeList2 = ee.List([5,6,7,8])
    >>> eeList3 = eeList1 + eeList2
    >>> eeList3.getInfo()
    [1,2,3,4,5,6,7,8]
    """
    return self.cat(other)


@extend(ee.ee_list.List)
def __radd__(self, other):
    """Concatenates the contents of a list onto another list.

    Parameters
    ----------
    self : ee.List
        Right operand.
    other : ee.List | list
        Left operand.

    Returns
    -------
    ee.List
        Concatenation of two lists.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> eeList = ee.List([1,2,3,4])
    >>> normalList = [5,6,7,8]
    >>> eeList2 = normalList + eeList
    >>> eeList2.getInfo()
    [5,6,7,8,1,2,3,4]
    """
    return ee.List(other).cat(self)


@extend(ee.ee_list.List)
def __mul__(self, other):
    """Returns a new list containing a list repeated n times.

    Parameters
    ----------
    self : ee.List
        Left operand.
    other : ee.Number | numeric
        Right operand.

    Returns
    -------
    ee.List
        Repeated list.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> eeList1 = ee.List([1,2,3,4])
    >>> eeList2 = eeList * 2
    >>> eeList2.getInfo()
    [1,2,3,4,1,2,3,4]
    """
    return ee.List.repeat(self, other).flatten()


@extend(ee.ee_list.List)
def __rmul__(self, other):
    """Returns a new list containing a list repeated n times.

    Parameters
    ----------
    self : ee.List
        Right operand.
    other : ee.Number | numeric
        Left operand.

    Returns
    -------
    ee.List
        Repeated list.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> eeList1 = ee.List([1,2,3,4])
    >>> eeList2 = 2 * eeList1
    >>> eeList2.getInfo()
    [1,2,3,4,1,2,3,4]
    """
    return ee.List.repeat(self, other).flatten()



def extend(cls, static=False):
    """Extends the cls class.

    Parameters
    ----------
    cls : class
        Class to extend.
    static : boolean, default = False
        Whether extend as a static method.

    Returns
    -------
    decorator
        Decorator for extending classes.
    """
    if static:
        return lambda f: (setattr(cls, f.__name__, staticmethod(f)) or f)
    else:
        return lambda f: (setattr(cls, f.__name__, f) or f)



"""The eemont python package."""

__version__ = "0.3.5"

from . import (app, dataframe, eeDictionary, eeList, extra, feature,
               featurecollection, geometry, image, imagecollection, number)
from .common import *



import warnings

import ee

from .extending import extend


@extend(ee.ee_number.Number)
def __add__(self, other):
    """Computes the addition between two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Addition of two numbers.
    """
    return self.add(other)


@extend(ee.ee_number.Number)
def __radd__(self, other):
    """Computes the addition between two numbers.

    Parameters
    ----------
    self : ee.Number
        Right operand.
    other : ee.Number | numeric
        Left operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Addition of two numbers.
    """
    return self.add(other)


@extend(ee.ee_number.Number)
def __sub__(self, other):
    """Computes the subtraction between two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Subtraction of two numbers.
    """
    return self.subtract(other)


@extend(ee.ee_number.Number)
def __rsub__(self, other):
    """Computes the subtraction between two numbers.

    Parameters
    ----------
    self : ee.Number
        Right operand.
    other : ee.Number | numeric
        Left operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Subtraction of two numbers.
    """
    return ee.Number(other).subtract(self)


@extend(ee.ee_number.Number)
def __mul__(self, other):
    """Computes the multiplication between two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Multiplication of two numbers.
    """
    return self.multiply(other)


@extend(ee.ee_number.Number)
def __rmul__(self, other):
    """Computes the multiplication between two numbers.

    Parameters
    ----------
    self : ee.Number
        Right operand.
    other : ee.Number | numeric
        Left operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Multiplication of two numbers.
    """
    return self.multiply(other)


@extend(ee.ee_number.Number)
def __truediv__(self, other):
    """Computes the division between two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Division of two numbers.
    """
    return self.divide(other)


@extend(ee.ee_number.Number)
def __rtruediv__(self, other):
    """Computes the division between two numbers.

    Parameters
    ----------
    self : ee.Number
        Right operand.
    other : ee.Number | numeric
        Left operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Division of two numbers.
    """
    return ee.Number(other).divide(self)


@extend(ee.ee_number.Number)
def __floordiv__(self, other):
    """Computes the floor division of two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Floor division of two numbers.
    """
    return self.divide(other).floor()


@extend(ee.ee_number.Number)
def __rfloordiv__(self, other):
    """Computes the floor division of two numbers.

    Parameters
    ----------
    self : ee.Number
        Right operand.
    other : ee.Number | numeric
        Left operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Floor division of two numbers.
    """
    return ee.Number(other).divide(self).floor()


@extend(ee.ee_number.Number)
def __mod__(self, other):
    """Computes the modulo of two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Modulo of two numbers.
    """
    return self.mod(other)


@extend(ee.ee_number.Number)
def __rmod__(self, other):
    """Computes the modulo of two numbers.

    Parameters
    ----------
    self : ee.Number
        Right operand.
    other : ee.Number | numeric
        Left operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Modulo of two numbers.
    """
    return ee.Number(other).mod(self)


@extend(ee.ee_number.Number)
def __pow__(self, other):
    """Computes the base (left operand) to the power (right operand).

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Bsae to the power of two numbers.
    """
    return self.pow(other)


@extend(ee.ee_number.Number)
def __rpow__(self, other):
    """Computes the base (left operand) to the power (right operand).

    Parameters
    ----------
    self : ee.Number
        Right operand.
    other : ee.Number | numeric
        Left operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Base to the power of two numbers.
    """
    return ee.Number(other).pow(self)


@extend(ee.ee_number.Number)
def __lshift__(self, other):
    """Computes the left shift operation between two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Left shift operation.
    """
    return self.leftShift(other)


@extend(ee.ee_number.Number)
def __rlshift__(self, other):
    """Computes the left shift operation between two numbers.

    Parameters
    ----------
    self : ee.Number
        Right operand.
    other : ee.Number | numeric
        Left operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Left shift operation.
    """
    return ee.Number(other).leftShift(self)


@extend(ee.ee_number.Number)
def __rshift__(self, other):
    """Computes the right shift operation between two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Right shift operation.
    """
    return self.rightShift(other)


@extend(ee.ee_number.Number)
def __rrshift__(self, other):
    """Computes the right shift operation between two numbers.

    Parameters
    ----------
    self : ee.Number
        Right operand.
    other : ee.Number | numeric
        Left operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Right shift operation.
    """
    return ee.Number(other).rightShift(self)


@extend(ee.ee_number.Number)
def __and__(self, other):
    """Computes the binary operator AND between two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Binary operator AND.
    """
    return self.And(other)


@extend(ee.ee_number.Number)
def __rand__(self, other):
    """Computes the binary operator AND between two numbers.

    Parameters
    ----------
    self : ee.Number
        Right operand.
    other : ee.Number | numeric
        Left operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Binary operator AND.
    """
    return ee.Number(other).And(self)


@extend(ee.ee_number.Number)
def __or__(self, other):
    """Computes the binary operator OR between two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Binary operator OR.
    """
    return self.Or(other)


@extend(ee.ee_number.Number)
def __ror__(self, other):
    """Computes the binary operator OR between two numbers.

    Parameters
    ----------
    self : ee.Number
        Right operand.
    other : ee.Number | numeric
        Left operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Binary operator OR.
    """
    return ee.Number(other).Or(self)


@extend(ee.ee_number.Number)
def __lt__(self, other):
    """Computes the rich comparison LOWER THAN between two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Rich comparison LOWER THAN.
    """
    return self.lt(other)


@extend(ee.ee_number.Number)
def __le__(self, other):
    """Computes the rich comparison LOWER THAN OR EQUAL between two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Rich comparison LOWER THAN OR EQUAL.
    """
    return self.lte(other)


@extend(ee.ee_number.Number)
def __eq__(self, other):
    """Computes the rich comparison EQUAL between two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Rich comparison EQUAL.
    """
    return self.eq(other)


@extend(ee.ee_number.Number)
def __ne__(self, other):
    """Computes the rich comparison NOT EQUAL THAN between two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Rich comparison NOT EQUAL.
    """
    return self.neq(other)


@extend(ee.ee_number.Number)
def __gt__(self, other):
    """Computes the rich comparison GREATER THAN between two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Rich comparison GREATER THAN.
    """
    return self.gt(other)


@extend(ee.ee_number.Number)
def __ge__(self, other):
    """Computes the rich comparison GREATER THAN OR EQUAL between two numbers.

    Parameters
    ----------
    self : ee.Number
        Left operand.
    other : ee.Number | numeric
        Right operand. If numeric, an ee.Number is created from its value.

    Returns
    -------
    ee.Number
        Rich comparison GREATER THAN OR EQUAL.
    """
    return self.gte(other)


@extend(ee.ee_number.Number)
def __neg__(self):
    """Computes the unary operator NEGATIVE on an image.

    Parameters
    ----------
    self : ee.Number
        Operand.

    Returns
    -------
    ee.Number
        Unary operator NEGATIVE.
    """
    return self.multiply(-1)


@extend(ee.ee_number.Number)
def __invert__(self):
    """Computes the unary operator NOT on an image.

    Parameters
    ----------
    self : ee.Number
        Operand.

    Returns
    -------
    ee.Number
        Unary operator NOT.
    """
    return self.Not()



import webbrowser

import ee
import re
import requests
from box import Box
from bs4 import BeautifulSoup
from ee_extra.Apps.core import apps as extra_apps

from .extending import extend


@extend(ee)
class App:
    """Google Earth Engine App Manager and Descriptor.

    Inspect, open and download Google Earth Engine apps.

    Parameters
    ----------
    url : str
        URL of the Google Earth Engine App.

    Examples
    --------
    >>> import ee, eemont
    >>> app = ee.App("https://jstnbraaten.users.earthengine.app/view/conus-cover-vis")
    >>> app.name
    'conus-cover-vis'
    >>> app.creator
    'jstnbraaten'
    >>> app.open()
    >>> app.download()
    """

    def __init__(self, url):
        if "users.earthengine.app" not in url:
            raise Exception("Not a valid Earth Engine App! Check the url again!")
        if "users.earthengine.app/view/" not in url:
            raise Exception(
                "This seems to be an Earth Engine App Collection! Please use the url of an app!"
            )

        self.url = url
        """URL of the App."""

        self.name = url.split("/")[4]
        """Name of the App."""

        self.creator = url.split("/")[2].split(".")[0]
        """Username of the App creator."""

        self.username = self.creator
        """Instance of creator."""

    def __repr__(self):
        return f"Google Earth Engine App (name: {self.name}, creator: {self.creator})"

    def open(self):
        """Opens the current app in the default browser if possible.

        Returns
        -------
        None

        Examples
        --------
        >>> import ee, eemont
        >>> app = ee.App("https://jstnbraaten.users.earthengine.app/view/conus-cover-vis")
        >>> app.open()
        """
        webbrowser.open(self.url)

    def download(self, file=None):
        """Downloads the current app to a JS file.

        Parameters
        ----------
        file : str, default = None
            Path + filename to store the file. If None, the name of the app is used.

        Returns
        -------
        None

        Examples
        --------
        >>> import ee, eemont
        >>> app = ee.App("https://jstnbraaten.users.earthengine.app/view/conus-cover-vis")
        >>> app.download()
        """
        r = requests.get(self.url)
        r = BeautifulSoup(r.text)
        r = r.find_all("main")[0].find("script").string
        app_url = re.findall(r"init\((.*?)\)", r)[0].replace('"', "").replace("'", "")
        r = requests.get(app_url).json()
        path = r["path"]
        r = r["dependencies"][path]
        if file is None:
            file = self.name + ".js"
        with open(file, "w") as fp:
            fp.write(r)


@extend(ee)
def listApps(online=False):
    """Gets the dictionary of available Google Earth Engine Apps from ee-appshot [1]_.

    Parameters
    ----------
    online : boolean
        Whether to retrieve the most recent list of apps directly from the GitHub
        repository and not from the local copy.

    Returns
    -------
    Box
        Dictionary of available apps.

    Examples
    --------
    >>> import ee, eemont
    >>> apps = ee.listApps()
    >>> apps.jstnbraaten.conus_cover_vis
    Google Earth Engine App (name: conus-cover-vis, creator: jstnbraaten)
    >>> apps.jstnbraaten.conus_cover_vis.name
    'conus-cover-vis'
    >>> apps.jstnbraaten.conus_cover_vis.open()
    >>> apps.jstnbraaten.conus_cover_vis.download()

    References
    ----------
    .. [1] Roy, Samapriya. 2021. ee-appshot: Create Snapshot of Earth Engine Apps.
        https://github.com/samapriya/ee-appshot
    """

    class GoogleEarthEngineApps(Box):
        def __repr__(self):
            return f"Google Earth Engine User Apps (items: {list(self.keys())})"

    ee_apps = extra_apps()
    apps_extra = {}
    for user, eeapps in ee_apps.items():
        apps_extra[user] = {}
        user_apps = []
        for app in eeapps:
            if "users.earthengine.app/view/" in app:
                the_app = App(app)
            apps_extra[user][the_app.name] = the_app

    return GoogleEarthEngineApps(apps_extra, frozen_box=True)



import io
import os
import re

from setuptools import find_packages, setup


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type(u"")
    with io.open(filename, mode="r", encoding="utf-8") as fd:
        return re.sub(text_type(r":[a-z]+:`~?(.*?)`"), text_type(r"``\1``"), fd.read())


setup(
    name="eemont",
    version="0.3.5",
    url="https://github.com/davemlz/eemont",
    license="MIT",
    author="David Montero Loaiza",
    author_email="dml.mont@gmail.com",
    description="A Python package that extends Google Earth Engine",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("tests",), include=["eemont", "eemont.*"]),
    package_data={"eemont": ["data/*.json"]},
    install_requires=[
        "beautifulsoup4",
        "earthengine-api",
        "ee_extra>=0.0.14",
        "numpy",
        "pandas",
        "python-box>=6.0.0",
        "geopy",
        "requests",
    ],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



import warnings

import ee
import ee_extra
import ee_extra.Spectral.core
import ee_extra.Algorithms.core
import numpy as np
import requests

from .extending import extend


@extend(ee.imagecollection.ImageCollection)
def __len__(self):
    """Returns the size of the image collection.

    Parameters
    ----------
    self : ee.ImageCollection
        Image Collection to get the size from.

    Returns
    -------
    int
        Size of the image collection.
    """
    return self.size().getInfo()


@extend(ee.imagecollection.ImageCollection)
def __getitem__(self, key):
    """Gets the band of each image in the image collection according to the specified key.

    Parameters
    ----------
    self : ee.ImageCollection
        Image Collection to get the bands from.
    key : numeric | string | list[numeric] | list[string] | slice
        Key used to get the specified band. If numeric, it gets the band at that index.
        If string, it gets the band with that name or that matches with regex. If list,
        it gets multiple bands. If slice, it calls the slice() method (the step parameter
        is ignored).

    Returns
    -------
    ee.ImageCollection
        Image Collection with the selected bands.
    """
    if isinstance(key, slice):

        if key.start == None:
            start = 0
        else:
            start = key.start

        if key.stop == None:
            stop = self.first().bandNames().size()
        else:
            stop = key.stop

        def sliceCollection(img):
            return img.slice(start, stop)

        selected = self.map(sliceCollection)

    else:
        selected = self.select(key)

    return selected


@extend(ee.imagecollection.ImageCollection)
def closest(self, date, tolerance=1, unit="month"):
    """Gets the closest image (or set of images if the collection intersects a region that
    requires multiple scenes) to the specified date.

    Tip
    ----------
    Check more info about getting the closest image to a specific date in the
    :ref:`User Guide<Closest Image to a Specific Date>`.

    Parameters
    ----------
    self : ee.ImageCollection [this]
        Image Collection from which to get the closest image to the specified date.
    date : ee.Date | string
        Date of interest. The method will look for images closest to this date.
    tolerance : float, default = 1
        Filter the collection to [date - tolerance, date + tolerance) before searching the
        closest image. This speeds up the searching process for collections
        with a high temporal resolution.
    unit : string, default = 'month'
        Units for tolerance. Available units: 'year', 'month', 'week', 'day', 'hour',
        'minute' or 'second'.

    Returns
    -------
    ee.ImageCollection
        Closest images to the specified date.

    Examples
    --------

    .. tabs::

        .. code-tab:: py

            import ee, eemont
            ee.Initialize()
            S2 = ee.ImageCollection('COPERNICUS/S2_SR').closest('2020-10-15')

        .. code-tab:: r R

            library(rgee)
            library(reticulate)
            ee_Initialize()
            eemont <- import("eemont")
            S2 <- ee$ImageCollection('COPERNICUS/S2_SR')$closest('2020-10-15')

        .. code-tab:: julia

            using EarthEngine, PyCall
            Initialize()
            eemont = pyimport("eemont");
            S2 = EE.ImageCollection("COPERNICUS/S2_SR") |> x -> closest(x,"2020-10-15");
    """
    return ee_extra.ImageCollection.core.closest(self, date, tolerance, unit)


@extend(ee.imagecollection.ImageCollection)
def getTimeSeriesByRegion(
    self,
    reducer,
    bands=None,
    geometry=None,
    scale=None,
    crs=None,
    crsTransform=None,
    bestEffort=False,
    maxPixels=1e12,
    tileScale=1,
    dateColumn="date",
    dateFormat="ISO",
    naValue=-9999,
):
    """Gets the time series by region for the given image collection and geometry (feature
    or feature collection are also supported) according to the specified reducer (or
    reducers).

    Tip
    ----------
    Check more info about time series in the :ref:`User Guide<Time Series By Regions>`.

    Parameters
    ----------
    self : ee.ImageCollection (this)
        Image collection to get the time series from.
    reducer : ee.Reducer | list[ee.Reducer]
        Reducer or list of reducers to use for region reduction.
    bands : str | list[str], default = None
        Selection of bands to get the time series from. Defaults to all bands in the image
        collection.
    geometry : ee.Geometry | ee.Feature | ee.FeatureCollection, default = None
        Geometry to perform the region reduction. If ee.Feature or ee.FeatureCollection,
        the geometry() method is called. In order to get reductions by each feature please
        see the getTimeSeriesByRegions() method. Defaults to the footprint of the first
        band for each image in the collection.
    scale : numeric, default = None
        Nomical scale in meters.
    crs : Projection, default = None
        The projection to work in. If unspecified, the projection of the image's first
        band is used. If specified in addition to scale, rescaled to the specified scale.
    crsTransform : list, default = None
        The list of CRS transform values. This is a row-major ordering of the 3x2
        transform matrix. This option is mutually exclusive with 'scale', and replaces any
        transform already set on the projection.
    bestEffort : boolean, default = False
        If the polygon would contain too many pixels at the given scale, compute and use a
        larger scale which would allow the operation to succeed.
    maxPixels : numeric, default = 1e12
        The maximum number of pixels to reduce.
    tileScale : numeric, default = 1
        A scaling factor used to reduce aggregation tile size; using a larger tileScale
        (e.g. 2 or 4) may enable computations that run out of memory with the default.
    dateColumn : str, default = 'date'
        Output name of the date column.
    dateFormat : str, default = 'ISO'
        Output format of the date column. Defaults to ISO. Available options: 'ms' (for
        milliseconds), 'ISO' (for ISO Standard Format) or a custom format pattern.
    naValue : numeric, default = -9999
        Value to use as NA when the region reduction doesn't retrieve a value due to
        masked pixels.

    Returns
    -------
    ee.FeatureCollection
        Time series by region retrieved as a Feature Collection.

    See Also
    --------
    getTimeSeriesByRegions : Gets the time series by regions for the given image
        collection and feature collection according to the specified reducer (or reducers).

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Initialize()
    >>> f1 = ee.Feature(ee.Geometry.Point([3.984770,48.767221]).buffer(50),{'ID':'A'})
    >>> f2 = ee.Feature(ee.Geometry.Point([4.101367,48.748076]).buffer(50),{'ID':'B'})
    >>> fc = ee.FeatureCollection([f1,f2])
    >>> S2 = (ee.ImageCollection('COPERNICUS/S2_SR')
    ...      .filterBounds(fc)
    ...      .filterDate('2020-01-01','2021-01-01')
    ...      .maskClouds()
    ...      .scaleAndOffset()
    ...      .spectralIndices(['EVI','NDVI']))
    >>> ts = S2.getTimeSeriesByRegion(reducer = [ee.Reducer.mean(),ee.Reducer.median()],
    ...                               geometry = fc,
    ...                               bands = ['EVI','NDVI'],
    ...                               scale = 10)
    """
    return ee_extra.TimeSeries.core.getTimeSeriesByRegion(
        self,
        reducer,
        bands,
        geometry,
        scale,
        crs,
        crsTransform,
        bestEffort,
        maxPixels,
        tileScale,
        dateColumn,
        dateFormat,
        naValue,
    )


@extend(ee.imagecollection.ImageCollection)
def getTimeSeriesByRegions(
    self,
    reducer,
    collection,
    bands=None,
    scale=None,
    crs=None,
    crsTransform=None,
    tileScale=1,
    dateColumn="date",
    dateFormat="ISO",
    naValue=-9999,
):
    """Gets the time series by regions for the given image collection and feature
    collection according to the specified reducer (or reducers).

    Tip
    ----------
    Check more info about time series in the :ref:`User Guide<Time Series By Regions>`.

    Parameters
    ----------
    self : ee.ImageCollection (this)
        Image collection to get the time series from.
    reducer : ee.Reducer | list[ee.Reducer]
        Reducer or list of reducers to use for region reduction.
    collection : ee.FeatureCollection
        Feature Collection to perform the reductions on. Image reductions are applied to
        each feature in the collection.
    bands : str | list[str], default = None
        Selection of bands to get the time series from. Defaults to all bands in the image
        collection.
    scale : numeric, default = None
        Nomical scale in meters.
    crs : Projection, default = None
        The projection to work in. If unspecified, the projection of the image's first
        band is used. If specified in addition to scale, rescaled to the specified scale.
    crsTransform : list, default = None
        The list of CRS transform values. This is a row-major ordering of the 3x2
        transform matrix. This option is mutually exclusive with 'scale', and replaces
        any transform already set on the projection.
    tileScale : numeric, default = 1
        A scaling factor used to reduce aggregation tile size; using a larger tileScale
        (e.g. 2 or 4) may enable computations that run out of memory with the default.
    dateColumn : str, default = 'date'
        Output name of the date column.
    dateFormat : str, default = 'ISO'
        Output format of the date column. Defaults to ISO. Available options: 'ms' (for
        milliseconds), 'ISO' (for ISO Standard Format) or a custom format pattern.
    naValue : numeric, default = -9999
        Value to use as NA when the region reduction doesn't retrieve a value due to
        masked pixels.

    Returns
    -------
    ee.FeatureCollection
        Time series by regions retrieved as a Feature Collection.

    See Also
    --------
    getTimeSeriesByRegion : Gets the time series by region for the given image collection
        and geometry (feature or feature collection are also supported)
        according to the specified reducer (or reducers).

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Initialize()
    >>> f1 = ee.Feature(ee.Geometry.Point([3.984770,48.767221]).buffer(50),{'ID':'A'})
    >>> f2 = ee.Feature(ee.Geometry.Point([4.101367,48.748076]).buffer(50),{'ID':'B'})
    >>> fc = ee.FeatureCollection([f1,f2])
    >>> S2 = (ee.ImageCollection('COPERNICUS/S2_SR')
    ...      .filterBounds(fc)
    ...      .filterDate('2020-01-01','2021-01-01')
    ...      .maskClouds()
    ...      .scaleAndOffset()
    ...      .spectralIndices(['EVI','NDVI']))
    >>> ts = S2.getTimeSeriesByRegions(reducer = [ee.Reducer.mean(),ee.Reducer.median()],
    ...                                collection = fc,
    ...                                bands = ['EVI','NDVI'],
    ...                                scale = 10)
    """
    return ee_extra.TimeSeries.core.getTimeSeriesByRegions(
        self,
        reducer,
        collection,
        bands,
        scale,
        crs,
        crsTransform,
        tileScale,
        dateColumn,
        dateFormat,
        naValue,
    )


@extend(ee.imagecollection.ImageCollection)
def index(
    self,
    index="NDVI",
    G=2.5,
    C1=6.0,
    C2=7.5,
    L=1.0,
    cexp=1.16,
    nexp=2.0,
    alpha=0.1,
    slope=1.0,
    intercept=0.0,
    gamma=1.0,
    kernel="RBF",
    sigma="0.5 * (a + b)",
    p=2.0,
    c=1.0,
    online=False,
    drop=False,
):
    """Computes one or more spectral indices (indices are added as bands) for an image
    collection.

    .. deprecated:: 0.3.0
       Use :func:`spectralIndices()` instead.

    Tip
    ----------
    Check more info about the supported platforms and spectral indices in the
    :ref:`User Guide<Spectral Indices Computation>`.

    Parameters
    ----------
    self : ee.ImageCollection
        Image collection to compute indices on. Must be scaled to [0,1].
    index : string | list[string], default = 'NDVI'
        Index or list of indices to compute.\n
        Available options:
            - 'vegetation' : Compute all vegetation indices.
            - 'burn' : Compute all burn indices.
            - 'water' : Compute all water indices.
            - 'snow' : Compute all snow indices.
            - 'drought' : Compute all drought indices.
            - 'urban' : Compute all urban (built-up) indices.
            - 'kernel' : Compute all kernel indices.
            - 'all' : Compute all indices listed below.
        Awesome Spectral Indices for GEE:
            Check the complete list of indices
            `here <https://awesome-ee-spectral-indices.readthedocs.io/en/latest/list.html>`_.
    G : float, default = 2.5
        Gain factor. Used just for index = 'EVI'.
    C1 : float, default = 6.0
        Coefficient 1 for the aerosol resistance term. Used just for index = 'EVI'.
    C2 : float, default = 7.5
        Coefficient 2 for the aerosol resistance term. Used just for index = 'EVI'.
    L : float, default = 1.0
        Canopy background adjustment. Used just for index = ['EVI','SAVI'].
    cexp : float, default = 1.16
        Exponent used for OCVI.
    nexp : float, default = 2.0
        Exponent used for GDVI.
    alpha : float, default = 0.1
        Weighting coefficient  used for WDRVI.
    slope : float, default = 1.0
        Soil line slope.
    intercept : float, default = 0.0
        Soil line intercept.
    gamma : float, default = 1.0
        Weighting coefficient  used for ARVI.
    kernel : str, default = 'RBF'
        Kernel used for kernel indices.\n
        Available options:
            - 'linear' : Linear Kernel.
            - 'RBF' : Radial Basis Function (RBF) Kernel.
            - 'poly' : Polynomial Kernel.
    sigma : str | float, default = '0.5 * (a + b)'
        Length-scale parameter. Used for kernel = 'RBF'. If str, this must be an
        expression including 'a' and 'b'. If numeric, this must be positive.
    p : float, default = 2.0
        Kernel degree. Used for kernel = 'poly'.
    c : float, default = 1.0
        Free parameter that trades off the influence of higher-order versus lower-order
        terms in the polynomial kernel. Used for kernel = 'poly'. This must be greater
        than or equal to 0.
    online : boolean, default = False
        Whether to retrieve the most recent list of indices directly from the GitHub
        repository and not from the local copy.
    drop : boolean, default = True
        Whether to drop all bands except the new spectral indices.

    Returns
    -------
    ee.ImageCollection
        Image collection with the computed spectral index, or indices, as new bands.

    See Also
    --------
    scale : Scales bands on an image collection.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> S2 = ee.ImageCollection('COPERNICUS/S2_SR').scale()

    - Computing one spectral index:

    >>> S2.index('NDVI')

    - Computing indices with different parameters:

    >>> S2.index('SAVI',L = 0.5)

    - Computing multiple indices:

    >>> S2.index(['NDVI','EVI','GNDVI'])

    - Computing a specific group of indices:

    >>> S2.index('vegetation')

    - Computing kernel indices:

    >>> S2.index(['kNDVI'],kernel = 'poly',p = 5)

    - Computing all indices:

    >>> S2.index('all')

    References
    ----------
    .. [1] https://awesome-ee-spectral-indices.readthedocs.io/en/latest/list.html
    """
    warnings.warn(
        "index() is deprecated, please use spectralIndices() instead",
        DeprecationWarning,
    )

    return self


@extend(ee.imagecollection.ImageCollection)
def spectralIndices(
    self,
    index="NDVI",
    G=2.5,
    C1=6.0,
    C2=7.5,
    L=1.0,
    cexp=1.16,
    nexp=2.0,
    alpha=0.1,
    slope=1.0,
    intercept=0.0,
    gamma=1.0,
    omega=2.0,
    beta=0.05,
    k=0.0,
    kernel="RBF",
    sigma="0.5 * (a + b)",
    p=2.0,
    c=1.0,
    lambdaN=858.5,
    lambdaR=645.0,
    lambdaG=555.0,
    online=False,
    drop=False,
):
    """Computes one or more spectral indices (indices are added as bands) for an image
    collection from the Awesome List of Spectral Indices.

    Tip
    ----------
    Check more info about the supported platforms and spectral indices in the
    :ref:`User Guide<Spectral Indices Computation>`.

    Parameters
    ----------
    self : ee.ImageCollection
        Image collection to compute indices on. Must be scaled to [0,1].
    index : string | list[string], default = 'NDVI'
        Index or list of indices to compute.\n
        Available options:
            - 'vegetation' : Compute all vegetation indices.
            - 'burn' : Compute all burn indices.
            - 'water' : Compute all water indices.
            - 'snow' : Compute all snow indices.
            - 'urban' : Compute all urban (built-up) indices.
            - 'kernel' : Compute all kernel indices.
            - 'all' : Compute all indices listed below.
        Awesome Spectral Indices for GEE:
            Check the complete list of indices
            `here <https://awesome-ee-spectral-indices.readthedocs.io/en/latest/list.html>`_.
    G : float, default = 2.5
        Gain factor. Used just for index = 'EVI'.
    C1 : float, default = 6.0
        Coefficient 1 for the aerosol resistance term. Used just for index = 'EVI'.
    C2 : float, default = 7.5
        Coefficient 2 for the aerosol resistance term. Used just for index = 'EVI'.
    L : float, default = 1.0
        Canopy background adjustment. Used just for index = ['EVI','SAVI'].
    cexp : float, default = 1.16
        Exponent used for OCVI.
    nexp : float, default = 2.0
        Exponent used for GDVI.
    alpha : float, default = 0.1
        Weighting coefficient  used for WDRVI.
    slope : float, default = 1.0
        Soil line slope.
    intercept : float, default = 0.0
        Soil line intercept.
    gamma : float, default = 1.0
        Weighting coefficient  used for ARVI.
    omega : float, default = 2.0
        Weighting coefficient  used for MBWI.
    beta : float, default = 0.05
        Calibration parameter used for NDSIns.
    k : float, default = 0.0
        Slope parameter by soil used for NIRvH2.
    kernel : str, default = 'RBF'
        Kernel used for kernel indices.\n
        Available options:
            - 'linear' : Linear Kernel.
            - 'RBF' : Radial Basis Function (RBF) Kernel.
            - 'poly' : Polynomial Kernel.
    sigma : str | float, default = '0.5 * (a + b)'
        Length-scale parameter. Used for kernel = 'RBF'. If str, this must be an
        expression including 'a' and 'b'. If numeric, this must be positive.
    p : float, default = 2.0
        Kernel degree. Used for kernel = 'poly'.
    c : float, default = 1.0
        Free parameter that trades off the influence of higher-order versus lower-order
        terms in the polynomial kernel. Used for kernel = 'poly'. This must be greater
        than or equal to 0.
    lambdaN : float, default = 858.5
        NIR wavelength used for NIRvH2 and NDGI.
    lambdaR : float, default = 645.0
        Red wavelength used for NIRvH2 and NDGI.
    lambdaG : float, default = 555.0
        Green wavelength used for NDGI.
    online : boolean, default = False
        Whether to retrieve the most recent list of indices directly from the GitHub
        repository and not from the local copy.
    drop : boolean, default = True
        Whether to drop all bands except the new spectral indices.

    Returns
    -------
    ee.ImageCollection
        Image collection with the computed spectral index, or indices, as new bands.

    See Also
    --------
    scaleAndOffset : Scales bands on an image collection.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> S2 = ee.ImageCollection('COPERNICUS/S2_SR').scaleAndOffset()

    - Computing one spectral index:

    >>> S2.spectralIndices('NDVI')

    - Computing indices with different parameters:

    >>> S2.spectralIndices('SAVI',L = 0.5)

    - Computing multiple indices:

    >>> S2.spectralIndices(['NDVI','EVI','GNDVI'])

    - Computing a specific group of indices:

    >>> S2.spectralIndices('vegetation')

    - Computing kernel indices:

    >>> S2.spectralIndices(['kNDVI'],kernel = 'poly',p = 5)

    - Computing all indices:

    >>> S2.spectralIndices('all')
    """
    return ee_extra.Spectral.core.spectralIndices(
        self,
        index,
        G,
        C1,
        C2,
        L,
        cexp,
        nexp,
        alpha,
        slope,
        intercept,
        gamma,
        omega,
        beta,
        k,
        kernel,
        sigma,
        p,
        c,
        lambdaN,
        lambdaR,
        lambdaG,
        online,
        drop,
    )


@extend(ee.imagecollection.ImageCollection)
def maskClouds(
    self,
    method="cloud_prob",
    prob=60,
    maskCirrus=True,
    maskShadows=True,
    scaledImage=False,
    dark=0.15,
    cloudDist=1000,
    buffer=250,
    cdi=None,
):
    """Masks clouds and shadows in an image collection (valid just for Surface
    Reflectance products).

    Tip
    ----------
    Check more info about the supported platforms and clouds masking in the
    :ref:`User Guide<Masking Clouds and Shadows>`.

    Parameters
    ----------
    self : ee.ImageCollection [this]
        Image collection to mask.
    method : string, default = 'cloud_prob'
        Method used to mask clouds.\n
        Available options:
            - 'cloud_prob' : Use cloud probability.
            - 'qa' : Use Quality Assessment band.
        This parameter is ignored for Landsat products.
    prob : numeric [0, 100], default = 60
        Cloud probability threshold. Valid just for method = 'cloud_prob'. This parameter
        is ignored for Landsat products.
    maskCirrus : boolean, default = True
        Whether to mask cirrus clouds. Valid just for method = 'qa'. This parameter is
        ignored for Landsat products.
    maskShadows : boolean, default = True
        Whether to mask cloud shadows. For more info see 'Braaten, J. 2020. Sentinel-2
        Cloud Masking with s2cloudless. Google Earth Engine, Community Tutorials'.
    scaledImage : boolean, default = False
        Whether the pixel values are scaled to the range [0,1] (reflectance values). This
        parameter is ignored for Landsat products.
    dark : float [0,1], default = 0.15
        NIR threshold. NIR values below this threshold are potential cloud shadows. This
        parameter is ignored for Landsat products.
    cloudDist : int, default = 1000
        Maximum distance in meters (m) to look for cloud shadows from cloud edges. This
        parameter is ignored for Landsat products.
    buffer : int, default = 250
        Distance in meters (m) to dilate cloud and cloud shadows objects. This parameter
        is ignored for Landsat products.
    cdi : float [-1,1], default = None
        Cloud Displacement Index threshold. Values below this threshold are considered
        potential clouds. A cdi = None means that the index is not used. For more info see
        'Frantz, D., HaS, E., Uhl, A., Stoffels, J., Hill, J. 2018. Improvement of the
        Fmask algorithm for Sentinel-2 images: Separating clouds from bright surfaces
        based on parallax effects. Remote Sensing of Environment 2015: 471-481'.
        This parameter is ignored for Landsat products.

    Returns
    -------
    ee.ImageCollection
        Cloud-shadow masked image collection.

    Notes
    -----
    This method may mask water as well as clouds for the Sentinel-3 Radiance product.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> S2 = (ee.ImageCollection('COPERNICUS/S2_SR')
    ...     .maskClouds(prob = 75,buffer = 300,cdi = -0.5))
    """
    return ee_extra.QA.clouds.maskClouds(
        self,
        method,
        prob,
        maskCirrus,
        maskShadows,
        scaledImage,
        dark,
        cloudDist,
        buffer,
        cdi,
    )


@extend(ee.imagecollection.ImageCollection)
def scale(self):
    """Scales bands on an image collection.

    .. deprecated:: 0.3.0
       Use :func:`scaleAndOffset()` instead.

    Tip
    ----------
    Check more info about the supported platforms and image scaling the
    :ref:`User Guide<Image Scaling>`.

    Parameters
    ----------
    self : ee.ImageCollection (this)
        Image collection to scale.

    Returns
    -------
    ee.ImageCollection
        Scaled image collection.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> S2 = ee.ImageCollection('COPERNICUS/S2_SR').scale()
    """
    warnings.warn(
        "scale() is deprecated, please use scaleAndOffset() instead",
        DeprecationWarning,
    )

    return ee_extra.STAC.core.scaleAndOffset(self)


@extend(ee.imagecollection.ImageCollection)
def getScaleParams(self):
    """Gets the scale parameters for each band of the image collection.

    Parameters
    ----------
    self : ee.ImageCollection (this)
        Image collection to get the scale parameters from.

    Returns
    -------
    dict
        Dictionary with the scale parameters for each band.

    See Also
    --------
    getOffsetParams : Gets the offset parameters for each band of the image collection.
    scaleAndOffset : Scales bands on an image collection according to their scale and
        offset parameters.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.ImageCollection('MODIS/006/MOD11A2').getScaleParams()
    {'Clear_sky_days': 1.0,
     'Clear_sky_nights': 1.0,
     'Day_view_angl': 1.0,
     'Day_view_time': 0.1,
     'Emis_31': 0.002,
     'Emis_32': 0.002,
     'LST_Day_1km': 0.02,
     'LST_Night_1km': 0.02,
     'Night_view_angl': 1.0,
     'Night_view_time': 0.1,
     'QC_Day': 1.0,
     'QC_Night': 1.0}
    """
    return ee_extra.STAC.core.getScaleParams(self)


@extend(ee.imagecollection.ImageCollection)
def getOffsetParams(self):
    """Gets the offset parameters for each band of the image collection.

    Parameters
    ----------
    self : ee.ImageCollection (this)
        Image collection to get the offset parameters from.

    Returns
    -------
    dict
        Dictionary with the offset parameters for each band.

    See Also
    --------
    getScaleParams : Gets the scale parameters for each band of the image collection.
    scaleAndOffset : Scales bands on an image collection according to their scale and
        offset parameters.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.ImageCollection('MODIS/006/MOD11A2').getOffsetParams()
    {'Clear_sky_days': 0.0,
     'Clear_sky_nights': 0.0,
     'Day_view_angl': -65.0,
     'Day_view_time': 0.0,
     'Emis_31': 0.49,
     'Emis_32': 0.49,
     'LST_Day_1km': 0.0,
     'LST_Night_1km': 0.0,
     'Night_view_angl': -65.0,
     'Night_view_time': 0.0,
     'QC_Day': 0.0,
     'QC_Night': 0.0}
    """
    return ee_extra.STAC.core.getOffsetParams(self)


@extend(ee.imagecollection.ImageCollection)
def scaleAndOffset(self):
    """Scales bands on an image collection according to their scale and offset parameters.

    Tip
    ----------
    Check more info about the supported platforms and image scaling the
    :ref:`User Guide<Image Scaling>`.

    Parameters
    ----------
    self : ee.ImageCollection (this)
        Image collection to scale.

    Returns
    -------
    ee.ImageCollection
        Scaled image collection.

    See Also
    --------
    getOffsetParams : Gets the offset parameters for each band of the image collection.
    getScaleParams : Gets the scale parameters for each band of the image collection.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> S2 = ee.ImageCollection('COPERNICUS/S2_SR').scaleAndOffset()
    """
    return ee_extra.STAC.core.scaleAndOffset(self)


@extend(ee.imagecollection.ImageCollection)
def preprocess(self, **kwargs):
    """Pre-processes the image collection: masks clouds and shadows, and scales and
    offsets the image collection.

    Tip
    ----------
    Check more info here about the supported platforms, :ref:`Image Scaling<Image Scaling>`
    and :ref:`Masking Clouds and Shadows<Masking Clouds and Shadows>`.

    Parameters
    ----------
    self : ee.ImageCollection [this]
        Image Collection to pre-process.
    **kwargs :
        Keywords arguments for maskClouds().

    Returns
    -------
    ee.ImageCollection
        Pre-processed image collection.

    See Also
    --------
    getScaleParams : Gets the scale parameters for each band of the image collection.
    getOffsetParams : Gets the offset parameters for each band of the image collection.
    scaleAndOffset : Scales bands on an image collection according to their scale and
        offset parameters.
    maskClouds : Masks clouds and shadows in an image collection.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> S2 = ee.ImageCollection('COPERNICUS/S2_SR').preprocess()
    """
    return ee_extra.QA.pipelines.preprocess(self, **kwargs)


@extend(ee.imagecollection.ImageCollection)
def getSTAC(self):
    """Gets the STAC of the image collection.

    Parameters
    ----------
    self : ee.ImageCollection [this]
        Image Collection to get the STAC from.

    Returns
    -------
    dict
        STAC of the image collection.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.ImageCollection('COPERNICUS/S2_SR').getSTAC()
    {'stac_version': '1.0.0-rc.2',
     'type': 'Collection',
     'stac_extensions': ['https://stac-extensions.github.io/eo/v1.0.0/schema.json'],
     'id': 'COPERNICUS/S2_SR',
     'title': 'Sentinel-2 MSI: MultiSpectral Instrument, Level-2A',
     'gee:type': 'image_collection',
     ...}
    """
    return ee_extra.STAC.core.getSTAC(self)


@extend(ee.imagecollection.ImageCollection)
def getDOI(self):
    """Gets the DOI of the image collection, if available.

    Parameters
    ----------
    self : ee.ImageCollection [this]
        Image Collection to get the DOI from.

    Returns
    -------
    str
        DOI of the ee.ImageCollection dataset.

    See Also
    --------
    getCitation : Gets the citation of the image collection, if available.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.ImageCollection('NASA/GPM_L3/IMERG_V06').getDOI()
    '10.5067/GPM/IMERG/3B-HH/06'
    """
    return ee_extra.STAC.core.getDOI(self)


@extend(ee.imagecollection.ImageCollection)
def getCitation(self):
    """Gets the citation of the image collection, if available.

    Parameters
    ----------
    self : ee.ImageCollection [this]
        Image Collection to get the citation from.

    Returns
    -------
    str
        Citation of the ee.ImageCollection dataset.

    See Also
    --------
    getDOI : Gets the DOI of the image collection, if available.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.ImageCollection('NASA/GPM_L3/IMERG_V06').getCitation()
    'Huffman, G.J., E.F. Stocker, D.T. Bolvin, E.J. Nelkin, Jackson Tan (2019),
    GPM IMERG Final Precipitation L3 Half Hourly 0.1 degree x 0.1 degree V06, Greenbelt,
    MD, Goddard Earth Sciences Data and Information Services Center (GES DISC),
    Accessed: [Data Access Date],
    [doi:10.5067/GPM/IMERG/3B-HH/06](https://doi.org/10.5067/GPM/IMERG/3B-HH/06)'
    """
    return ee_extra.STAC.core.getCitation(self)


@extend(ee.imagecollection.ImageCollection)
def panSharpen(self, method="SFIM", qa=None, **kwargs):
    """Apply panchromatic sharpening to each Image in the Image Collection.

    Optionally, run quality assessments between the original and sharpened Images to
    measure spectral distortion and set results as properties of each sharpened Image.

    Parameters
    ----------
    self : ee.ImageCollection [this]
        Image Collection to sharpen.
    method : str, default="SFIM"
        The sharpening algorithm to apply. Current options are "SFIM" (Smoothing
        Filter-based Intensity Modulation), "HPFA" (High Pass Filter Addition), "PCS"
        (Principal Component Substitution), and "SM" (simple mean). Different
        sharpening methods will produce different quality sharpening results in different
        scenarios.
    qa : str | list, default=None
        One or more optional quality assessment names to apply after sharpening. Results 
        will be stored as image properties with the pattern `eemont:metric`, e.g. `eemont:RMSE`.
    **kwargs :
        Keyword arguments passed to ee.Image.reduceRegion() such as "geometry",
        "maxPixels", "bestEffort", etc. These arguments are only used for PCS sharpening
        and quality assessments.

    Returns
    -------
    ee.ImageCollection
        The Image Collection with all sharpenable bands sharpened to the panchromatic
        resolution and quality assessments run and set as properties.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> source = ee.ImageCollection("LANDSAT/LC08/C01/T1_TOA")
    >>> sharp = source.panSharpen(method="HPFA", qa=["MSE", "RMSE"], maxPixels=1e13)
    """
    return ee_extra.Algorithms.core.panSharpen(img=self, method=method, qa=qa, prefix="eemont", **kwargs)


@extend(ee.imagecollection.ImageCollection)
def tasseledCap(self):
    """Calculates tasseled cap brightness, wetness, and greenness components for all
    images in the collection.

    Tasseled cap transformations are applied using coefficients published for these
    supported platforms:

    * Sentinel-2 MSI Level 1C [1]_
    * Landsat 9 OLI-2 SR [2]_
    * Landsat 9 OLI-2 TOA [2]_
    * Landsat 8 OLI SR [2]_
    * Landsat 8 OLI TOA [2]_
    * Landsat 7 ETM+ TOA [3]_
    * Landsat 5 TM Raw DN [4]_
    * Landsat 4 TM Raw DN [5]_
    * Landsat 4 TM Surface Reflectance [6]_
    * MODIS NBAR [7]_

    Parameters
    ----------
    self : ee.ImageCollection
        Image Collection to calculate tasseled cap components for. Must belong to a
        supported platform.

    Returns
    -------
    ee.ImageCollection
        Image Collection with the tasseled cap components as new bands in each image.

    References
    ----------
    .. [1] Shi, T., & Xu, H. (2019). Derivation of Tasseled Cap Transformation
        Coefficients for Sentinel-2 MSI At-Sensor Reflectance Data. IEEE Journal
        of Selected Topics in Applied Earth Observations and Remote Sensing, 1–11.
        doi:10.1109/jstars.2019.2938388
    .. [2] Zhai, Y., Roy, D.P., Martins, V.S., Zhang, H.K., Yan, L., Li, Z. 2022.
        Conterminous United States Landsat-8 top of atmosphere and surface reflectance
        tasseled cap transformation coefficeints. Remote Sensing of Environment, 
        274(2022). doi:10.1016/j.rse.2022.112992
    .. [3] Huang, C., Wylie, B., Yang, L., Homer, C. and Zylstra, G., 2002.
        Derivation of a tasselled cap transformation based on Landsat 7 at-satellite
        reflectance. International journal of remote sensing, 23(8), pp.1741-1748.
    .. [4] Crist, E.P., Laurin, R. and Cicone, R.C., 1986, September. Vegetation and
        soils information contained in transformed Thematic Mapper data. In
        Proceedings of IGARSS’86 symposium (pp. 1465-1470). Paris: European Space
        Agency Publications Division.
    .. [5] Crist, E.P. and Cicone, R.C., 1984. A physically-based transformation of
        Thematic Mapper data---The TM Tasseled Cap. IEEE Transactions on Geoscience
        and Remote sensing, (3), pp.256-263.
    .. [6] Crist, E.P., 1985. A TM tasseled cap equivalent transformation for
        reflectance factor data. Remote sensing of Environment, 17(3), pp.301-306.
    .. [7] Lobser, S.E. and Cohen, W.B., 2007. MODIS tasselled cap: land cover
        characteristics expressed through transformed MODIS data. International
        Journal of Remote Sensing, 28(22), pp.5079-5101.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> col = ee.ImageCollection("LANDSAT/LT05/C01/T1")
    >>> col = col.tasseledCap()
    """
    return ee_extra.Spectral.core.tasseledCap(self)



import warnings

import ee
import pandas as pd


def _extend_pdDataFrame():
    """Decorator. Extends the pd.DataFrame class."""
    return lambda f: (setattr(pd.core.frame.DataFrame, f.__name__, f) or f)


@_extend_pdDataFrame()
def toEEFeatureCollection(self, latitude=None, longitude=None):
    """Converts a pd.DataFrame object into an ee.FeatureCollection object.

    If lat/lon coordinates are available, the Data Frame can be converted into
    a Feature Collection with an associated geometry.

    Tip
    ----------
    Check more info about data conversion in the :ref:`User Guide<Data Conversion>`.

    Parameters
    ----------
    self : pd.DataFrame [this]
        Data Frame to convert into a Feature Collection.
    latitude : string
        Name of a latitude column, if available. Coupled with a longitude column,
        an ee.Geometry.Point is created and associated to each Feature.
    longitude : string
        Name of a longitude column, if available. Coupled with a latitude column,
        an ee.Geometry.Point is created and associated to each Feature.

    Returns
    -------
    ee.FeatureCollection
        Data Frame converted into a Feature Collection.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> df = pd.DataFrame()
    >>> df['lat'] = [2.92846, 4.8927]
    >>> df['lon'] = [-76.0269, -75.3188]
    >>> df['name'] = ['Nevado del Huila', 'Nevado del Ruiz']
    >>> fc = df.toEEFeatureCollection(latitude = 'lat',longitude = 'lon')
    """

    def getFeature(r):
        properties = r.to_dict()
        if latitude != None and longitude == None:
            warnings.warn(
                "longitude missing, Feature Collection with no geometries generated!",
                Warning,
            )
            return ee.Feature(None, properties)
        elif latitude == None and longitude != None:
            warnings.warn(
                "latitude missing, Feature Collection with no geometries generated!",
                Warning,
            )
            return ee.Feature(None, properties)
        elif latitude != None and longitude != None:
            point = ee.Geometry.Point([r[longitude], r[latitude]])
            return ee.Feature(point, properties)
        else:
            return ee.Feature(None, properties)

    dataFrame = self.copy()
    dataFrame["feature"] = dataFrame.apply(getFeature, axis=1)
    featureCollection = ee.FeatureCollection(dataFrame["feature"].tolist())

    return featureCollection



import ee
import geopy
from geopy.geocoders import get_geocoder_for_service

from .common import _lnglat_from_location, _retrieve_location
from .extending import extend
from .geometry import *


@extend(ee.feature.Feature)
def __getitem__(self, key):
    """Gets the column of the feature according to the specified key.

    Parameters
    ----------
    self : ee.Feature
        Feature to get the columns from.
    key : string | list[string]
        Key used to get the specified column. If string, it gets the column with that name
        or that matches with regex. If list, it gets multiple columns.

    Returns
    -------
    ee.Feature
        Feature with the selected columns.
    """
    if isinstance(key, str):
        key = [key]

    return self.select(key)


@extend(ee.feature.Feature, static=True)
def PointFromQuery(query, geocoder="nominatim", **kwargs):
    """Constructs an ee.Feature describing a point from a query submitted to a geodocer
    using the geopy package.

    This returns exactly one pair of coordinates. The properties of the feature correspond
    to the raw properties retrieved by the location of the query.

    Tip
    ----------
    Check more info about constructors in the :ref:`User Guide<Constructors>`.

    Parameters
    ----------
    query : str
        Address, query or structured query to geocode.
    geocoder : str, default = 'nominatim'
        Geocoder to use. Please visit https://geopy.readthedocs.io/ for more info.
    **kwargs :
        Keywords arguments for geolocator.geocode(). The user_agent argument is mandatory
        (this argument can be set as user_agent = 'my-gee-username' or user_agent =
        'my-gee-app-name'). Please visit https://geopy.readthedocs.io/ for more info.

    Returns
    -------
    ee.Feature
        Feature with a geometry describing a point from the specified query.

    See Also
    --------
    BBoxFromQuery : Constructs an ee.Feature describing a bounding box from a query
        submitted to a geodocer using the geopy package.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.Feature.PointFromQuery('Mt. Rainier, USA',
    ...                           user_agent = 'my-gee-eemont-query').getInfo()
    {'type': 'Feature',
     'geometry': {'type': 'Point', 'coordinates': [-121.757682, 46.8521484]},
     'properties': {'boundingbox': ['46.8520984',
       '46.8521984',
       '-121.757732',
       '-121.757632'],
      'class': 'natural',
      'display_name': 'Mount Rainier, Pierce County, Washington, United States',
      'importance': 0.5853390667167165,
      'lat': '46.8521484',
      'licence': 'Data © OpenStreetMap contributors, ODbL 1.0. https://osm.org/copyright',
      'lon': '-121.757682',
      'osm_id': 1744903493,
      'osm_type': 'node',
      'place_id': 17287419,
      'type': 'volcano'}}
    """
    location = _retrieve_location(query, geocoder, True, **kwargs)
    geometry = ee.Geometry.Point(_lnglat_from_location(location))

    return ee.Feature(geometry, location.raw)


@extend(ee.feature.Feature, static=True)
def BBoxFromQuery(query, geocoder="nominatim", **kwargs):
    """Constructs an ee.Feature describing a bounding box from a query submitted to a
    geodocer using the geopy package.

    The properties of the feature correspond to the raw properties retrieved by the
    location of the query.

    Tip
    ----------
    Check more info about constructors in the :ref:`User Guide<Constructors>`.

    Parameters
    ----------
    query : str
        Address, query or structured query to geocode.
    geocoder : str, default = 'nominatim'
        Geocoder to use. One of 'nominatim' or 'arcgis'. Please visit
        https://geopy.readthedocs.io/ for more info.
    **kwargs :
        Keywords arguments for geolocator.geocode(). The user_agent argument is mandatory
        (this argument can be set as user_agent = 'my-gee-username' or user_agent =
        'my-gee-app-name'). Please visit https://geopy.readthedocs.io/ for more info.

    Returns
    -------
    ee.Feature
        Feature with a geometry describing a bounding box from the specified query.

    See Also
    --------
    PointFromQuery : Constructs an ee.Feature describing a point from a query submitted to
        a geodocer using the geopy package.

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> ee.Feature.BBoxFromQuery('Bogotá',user_agent = 'my-gee-eemont-query').getInfo()
    {'type': 'Feature',
     'geometry': {'geodesic': False,
      'type': 'Polygon',
      'coordinates': [[[-74.22351370000001, 4.4711754],
        [-74.0102483, 4.4711754],
        [-74.0102483, 4.8331695],
        [-74.22351370000001, 4.8331695],
        [-74.22351370000001, 4.4711754]]]},
     'properties': {'boundingbox': ['4.4711754',
       '4.8331695',
       '-74.2235137',
       '-74.0102483'],
      'class': 'boundary',
      'display_name': 'Bogotá, Bogotá Distrito Capital, Región Andina, 11001, Colombia',
      'icon': 'https://nominatim.openstreetmap.org/ui/mapicons//poi_boundary_administrative.p.20.png',
      'importance': 0.7931743429157826,
      'lat': '4.6533326',
      'licence': 'Data © OpenStreetMap contributors, ODbL 1.0. https://osm.org/copyright',
      'lon': '-74.083652',
      'osm_id': 7426387,
      'osm_type': 'relation',
      'place_id': 259216862,
      'type': 'administrative'}}
    """
    if geocoder not in ["nominatim", "arcgis"]:
        raise Exception('Invalid geocoder! Use one of "nominatim" or "arcgis".')

    location = _retrieve_location(query, geocoder, True, **kwargs)

    if geocoder == "nominatim":
        BBox = location.raw["boundingbox"]
        geometry = ee.Geometry.BBox(
            float(BBox[2]), float(BBox[0]), float(BBox[3]), float(BBox[1])
        )
        return ee.Feature(geometry, location.raw)
    elif geocoder == "arcgis":
        BBox = location.raw["extent"]
        geometry = ee.Geometry.BBox(
            BBox["xmin"], BBox["ymin"], BBox["xmax"], BBox["ymax"]
        )
        return ee.Feature(geometry, location.raw)
    else:
        raise Exception('Invalid geocoder! Use one of "nominatim" or "arcgis".')


@extend(ee.feature.Feature, static=False)
def plusCodes(self, codeLength=10):
    """Converts the coordinates of the geometry of an ee.Feature to Plus Codes.

    Parameters
    ----------
    self : ee.Feature
        The feature to extract coordinates from.
    codeLength : int, default = 10
        The number of significant digits in the output codes, between 2 and 15. Shorter
        codes are less precise.

    Returns
    -------
    list | str
        The coordinates of the geometry converted to Plus Codes. The structure of the Plus
        Codes array will be identical to the structure returned by ee.Geometry.coordinates().

    Examples
    --------
    >>> import ee, eemont
    >>> ee.Authenticate()
    >>> ee.Initialize()
    >>> pt = ee.Feature(ee.Geometry.Point([-105, 40]))
    >>> pt.plusCodes()
    '85GQ2222+22'
    """
    return ee.Element.geometry(self).plusCodes(codeLength)
